import datetime
import os
import random
from typing import NamedTuple

import pytest
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec, ed25519, padding, rsa, x25519

from ykman.piv import (
    check_key,
    generate_csr,
    generate_self_signed_certificate,
    get_pivman_data,
    get_pivman_protected_data,
    pivman_set_mgm_key,
)
from ykman.util import parse_certificates, parse_private_key
from yubikit.core import TRANSPORT, NotSupportedError
from yubikit.core.smartcard import AID, ApduError
from yubikit.management import CAPABILITY, RELEASE_TYPE, ManagementSession
from yubikit.piv import (
    ALGORITHM,
    KEY_TYPE,
    MANAGEMENT_KEY_TYPE,
    OBJECT_ID,
    PIN_POLICY,
    SLOT,
    TOUCH_POLICY,
    InvalidPinError,
    PivSession,
    _do_check_key_support,
)

from ..util import open_file
from . import condition

DEFAULT_PIN = "123456"
NON_DEFAULT_PIN = "12341235"
DEFAULT_PUK = "12345678"
NON_DEFAULT_PUK = "12341236"
DEFAULT_MANAGEMENT_KEY = bytes.fromhex(
    "010203040506070801020304050607080102030405060708"
)
NON_DEFAULT_MANAGEMENT_KEY = bytes.fromhex(
    "010103040506070801020304050607080102030405060708"
)

NOW = datetime.datetime.now()

SIGN_KEY_TYPES = list(set(KEY_TYPE) - {KEY_TYPE.X25519})
ECDH_KEY_TYPES = [KEY_TYPE.ECCP256, KEY_TYPE.ECCP384, KEY_TYPE.X25519]


def get_test_cert():
    with open_file("rsa_2048_cert.pem") as f:
        return parse_certificates(f.read(), None)[0]


def get_test_key():
    with open_file("rsa_2048_key.pem") as f:
        return parse_private_key(f.read(), None)


@pytest.fixture
def scp(info, transport, scp_params):
    if transport == TRANSPORT.NFC and CAPABILITY.PIV in info.fips_capable:
        return scp_params
    return None


@pytest.fixture
@condition.capability(CAPABILITY.PIV)
def session(ccid_connection, scp, info):
    if CAPABILITY.PIV in info.reset_blocked:
        mgmt = ManagementSession(ccid_connection)
        mgmt.device_reset()
        piv = PivSession(ccid_connection, scp)
    else:
        piv = PivSession(ccid_connection, scp)
        piv.reset()
    yield piv
    reset_state(piv, scp)


class Keys(NamedTuple):
    pin: str
    puk: str
    mgmt: bytes


@pytest.fixture
def default_keys():
    yield Keys(DEFAULT_PIN, DEFAULT_PUK, DEFAULT_MANAGEMENT_KEY)


@pytest.fixture
def keys(session, info, default_keys, scp):
    if info.pin_complexity:
        new_keys = Keys(
            "12345679" if CAPABILITY.PIV in info.fips_capable else "123458",
            "12345670",
            bytes.fromhex("010203040506070801020304050607080102030405060709"),
        )
        session.change_pin(default_keys.pin, new_keys.pin)
        session.change_puk(default_keys.puk, new_keys.puk)
        session.authenticate(default_keys.mgmt)
        session.set_management_key(session.management_key_type, new_keys.mgmt)
        reset_state(session, scp)

        yield new_keys
    else:
        yield default_keys


def not_roca(version):
    return not ((4, 2, 0) <= version < (4, 3, 5))


def reset_state(session, scp_params):
    session.protocol.connection.connection.disconnect()
    session.protocol.connection.connection.connect()
    session.protocol.select(AID.PIV)
    if scp_params:
        session.protocol.init_scp(scp_params)


def assert_mgm_key_is(session, key):
    session.authenticate(key)


def assert_mgm_key_is_not(session, key):
    with pytest.raises(ApduError):
        session.authenticate(key)


def generate_key(
    session,
    scp,
    keys,
    slot=SLOT.AUTHENTICATION,
    key_type=KEY_TYPE.ECCP256,
    pin_policy=PIN_POLICY.DEFAULT,
):
    session.authenticate(keys.mgmt)
    key = session.generate_key(slot, key_type, pin_policy=pin_policy)
    reset_state(session, scp)
    return key


def generate_sw_key(key_type):
    if key_type.algorithm == ALGORITHM.RSA:
        return rsa.generate_private_key(65537, key_type.bit_len, default_backend())
    elif key_type == KEY_TYPE.ECCP256:
        return ec.generate_private_key(ec.SECP256R1(), default_backend())
    elif key_type == KEY_TYPE.ECCP384:
        return ec.generate_private_key(ec.SECP384R1(), default_backend())
    elif key_type == KEY_TYPE.ED25519:
        return ed25519.Ed25519PrivateKey.generate()
    elif key_type == KEY_TYPE.X25519:
        return x25519.X25519PrivateKey.generate()


def import_key(
    session,
    scp,
    keys,
    slot=SLOT.AUTHENTICATION,
    key_type=KEY_TYPE.ECCP256,
    pin_policy=PIN_POLICY.DEFAULT,
):
    private_key = generate_sw_key(key_type)
    session.authenticate(keys.mgmt)
    session.put_key(slot, private_key, pin_policy)
    reset_state(session, scp)
    return private_key.public_key()


def verify_cert_signature(cert, public_key=None):
    if not public_key:
        public_key = cert.public_key

    args = [cert.signature, cert.tbs_certificate_bytes, cert.signature_hash_algorithm]
    key_type = KEY_TYPE.from_public_key(public_key)
    if key_type.algorithm == ALGORITHM.RSA:
        args.insert(2, padding.PKCS1v15())
    elif key_type == KEY_TYPE.ED25519:
        args.pop()
    else:
        args[2] = ec.ECDSA(args[2])
    public_key.verify(*args)


def skip_unsupported_key_type(key_type, info, pin_policy=PIN_POLICY.DEFAULT):
    try:
        _do_check_key_support(
            info.version,
            key_type,
            pin_policy,
            TOUCH_POLICY.DEFAULT,
            fips_restrictions=CAPABILITY.PIV in info.fips_capable,
        )
    except NotSupportedError as e:
        pytest.skip(f"{e}")


class TestCertificateSignatures:
    @pytest.mark.parametrize("key_type", SIGN_KEY_TYPES)
    @pytest.mark.parametrize(
        "hash_algorithm", (hashes.SHA256, hashes.SHA384, hashes.SHA512)
    )
    def test_generate_self_signed_certificate(
        self, info, session, key_type, hash_algorithm, keys, scp
    ):
        skip_unsupported_key_type(key_type, info)

        slot = SLOT.SIGNATURE
        public_key = import_key(session, scp, keys, slot, key_type)
        session.authenticate(keys.mgmt)
        session.verify_pin(keys.pin)
        cert = generate_self_signed_certificate(
            session, slot, public_key, "CN=alice", NOW, NOW, hash_algorithm
        )

        if key_type == KEY_TYPE.ED25519:
            assert cert.public_key() == public_key
        else:
            assert cert.public_key().public_numbers() == public_key.public_numbers()
        verify_cert_signature(cert, public_key)


class TestDecrypt:
    @pytest.mark.parametrize(
        "key_type",
        [KEY_TYPE.RSA1024, KEY_TYPE.RSA2048, KEY_TYPE.RSA3072, KEY_TYPE.RSA4096],
    )
    def test_import_decrypt(self, session, info, key_type, keys, scp):
        skip_unsupported_key_type(key_type, info)

        public_key = import_key(
            session, scp, keys, SLOT.KEY_MANAGEMENT, key_type=key_type
        )
        pt = os.urandom(32)
        ct = public_key.encrypt(pt, padding.PKCS1v15())

        session.verify_pin(keys.pin)
        pt2 = session.decrypt(SLOT.KEY_MANAGEMENT, ct, padding.PKCS1v15())
        assert pt == pt2


class TestKeyAgreement:
    @pytest.mark.parametrize("key_type", ECDH_KEY_TYPES)
    def test_generate_ecdh(self, session, info, key_type, keys, scp):
        skip_unsupported_key_type(key_type, info)

        e_priv = generate_sw_key(key_type)
        public_key = generate_key(
            session, scp, keys, SLOT.KEY_MANAGEMENT, key_type=key_type
        )
        if key_type == KEY_TYPE.X25519:
            args = (public_key,)
        else:
            args = (ec.ECDH(), public_key)

        shared1 = e_priv.exchange(*args)
        session.verify_pin(keys.pin)
        shared2 = session.calculate_secret(SLOT.KEY_MANAGEMENT, e_priv.public_key())
        assert shared1 == shared2

    @pytest.mark.parametrize("key_type", ECDH_KEY_TYPES)
    def test_import_ecdh(self, session, info, key_type, keys, scp):
        skip_unsupported_key_type(key_type, info)

        e_priv = generate_sw_key(key_type)
        public_key = import_key(
            session, scp, keys, SLOT.KEY_MANAGEMENT, key_type=key_type
        )
        if key_type == KEY_TYPE.X25519:
            args = (public_key,)
        else:
            args = (ec.ECDH(), public_key)

        shared1 = e_priv.exchange(*args)
        session.verify_pin(keys.pin)
        shared2 = session.calculate_secret(SLOT.KEY_MANAGEMENT, e_priv.public_key())
        assert shared1 == shared2


class TestKeyManagement:
    def test_delete_certificate_requires_authentication(self, session, keys, scp):
        generate_key(session, scp, keys, SLOT.AUTHENTICATION)

        with pytest.raises(ApduError):
            session.delete_certificate(SLOT.AUTHENTICATION)

        session.authenticate(keys.mgmt)
        session.delete_certificate(SLOT.AUTHENTICATION)

    def test_generate_csr_works(self, session, keys, scp):
        public_key = generate_key(session, scp, keys, SLOT.AUTHENTICATION)

        session.verify_pin(keys.pin)
        csr = generate_csr(session, SLOT.AUTHENTICATION, public_key, "CN=alice")

        assert csr.public_key().public_numbers() == public_key.public_numbers()
        assert (
            csr.subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0].value
            == "alice"
        )

    def test_generate_self_signed_certificate_requires_pin(self, session, keys, scp):
        session.verify_pin(keys.pin)
        public_key = generate_key(session, scp, keys, SLOT.AUTHENTICATION)

        with pytest.raises(ApduError):
            generate_self_signed_certificate(
                session, SLOT.AUTHENTICATION, public_key, "CN=alice", NOW, NOW
            )

        session.verify_pin(keys.pin)
        generate_self_signed_certificate(
            session, SLOT.AUTHENTICATION, public_key, "CN=alice", NOW, NOW
        )

    @pytest.mark.parametrize("slot", (SLOT.SIGNATURE, SLOT.AUTHENTICATION))
    def test_generate_self_signed_certificate(self, session, slot, keys, scp):
        public_key = generate_key(session, scp, keys, slot)
        session.authenticate(keys.mgmt)
        session.verify_pin(keys.pin)
        cert = generate_self_signed_certificate(
            session, slot, public_key, "CN=alice", NOW, NOW
        )

        assert cert.public_key().public_numbers() == public_key.public_numbers()
        assert (
            cert.subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0].value
            == "alice"
        )

    def test_generate_key_requires_authentication(self, session, keys):
        with pytest.raises(ApduError):
            session.generate_key(
                SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, touch_policy=TOUCH_POLICY.DEFAULT
            )

        session.authenticate(keys.mgmt)
        session.generate_key(SLOT.AUTHENTICATION, KEY_TYPE.ECCP256)

    def test_put_certificate_requires_authentication(self, session, keys):
        cert = get_test_cert()
        with pytest.raises(ApduError):
            session.put_certificate(SLOT.AUTHENTICATION, cert)

        session.authenticate(keys.mgmt)
        session.put_certificate(SLOT.AUTHENTICATION, cert)

    def _test_put_key_pairing(self, session, scp, keys, alg1, alg2):
        # Set up a key in the slot and create a certificate for it
        public_key = generate_key(
            session, scp, keys, SLOT.AUTHENTICATION, key_type=alg1
        )
        session.authenticate(keys.mgmt)
        session.verify_pin(keys.pin)
        cert = generate_self_signed_certificate(
            session, SLOT.AUTHENTICATION, public_key, "CN=test", NOW, NOW
        )
        session.put_certificate(SLOT.AUTHENTICATION, cert)
        assert check_key(session, SLOT.AUTHENTICATION, cert.public_key())

        cert2 = session.get_certificate(SLOT.AUTHENTICATION)
        assert cert == cert2

        session.delete_certificate(SLOT.AUTHENTICATION)

        # Overwrite the key with one of the same type
        generate_key(session, scp, keys, SLOT.AUTHENTICATION, key_type=alg1)
        session.verify_pin(keys.pin)
        assert not check_key(session, SLOT.AUTHENTICATION, cert.public_key())

        # Overwrite the key with one of a different type
        generate_key(session, scp, keys, SLOT.AUTHENTICATION, key_type=alg2)
        session.verify_pin(keys.pin)
        assert not check_key(session, SLOT.AUTHENTICATION, cert.public_key())

    @condition.check(not_roca)
    @condition.yk4_fips(False)
    def test_put_certificate_verifies_key_pairing_rsa1024(
        self, session, keys, info, scp
    ):
        if CAPABILITY.PIV in info.fips_capable:
            pytest.skip("RSA1024 not available on YubiKey FIPS")
        self._test_put_key_pairing(
            session, scp, keys, KEY_TYPE.RSA1024, KEY_TYPE.ECCP256
        )

    @condition.check(not_roca)
    def test_put_certificate_verifies_key_pairing_rsa2048(self, session, keys, scp):
        self._test_put_key_pairing(
            session, scp, keys, KEY_TYPE.RSA2048, KEY_TYPE.ECCP256
        )

    @condition.check(not_roca)
    def test_put_certificate_verifies_key_pairing_eccp256_a(self, session, keys, scp):
        self._test_put_key_pairing(
            session, scp, keys, KEY_TYPE.ECCP256, KEY_TYPE.RSA2048
        )

    @condition.min_version(4)
    def test_put_certificate_verifies_key_pairing_eccp256_b(self, session, keys, scp):
        self._test_put_key_pairing(
            session, scp, keys, KEY_TYPE.ECCP256, KEY_TYPE.ECCP384
        )

    @condition.min_version(4)
    def test_put_certificate_verifies_key_pairing_eccp384(self, session, keys, scp):
        self._test_put_key_pairing(
            session, scp, keys, KEY_TYPE.ECCP384, KEY_TYPE.ECCP256
        )

    def test_put_key_requires_authentication(self, session, keys):
        private_key = get_test_key()
        with pytest.raises(ApduError):
            session.put_key(SLOT.AUTHENTICATION, private_key)

        session.authenticate(keys.mgmt)
        session.put_key(SLOT.AUTHENTICATION, private_key)

    def test_get_certificate_does_not_require_authentication(self, session, keys, scp):
        cert = get_test_cert()
        session.authenticate(keys.mgmt)
        session.put_certificate(SLOT.AUTHENTICATION, cert)
        reset_state(session, scp)

        assert session.get_certificate(SLOT.AUTHENTICATION)


class TestCompressedCertificate:
    def test_put_and_read_compressed_certificate(self, session, keys):
        session.authenticate(keys.mgmt)
        cert = get_test_cert()
        session.put_certificate(SLOT.AUTHENTICATION, cert)
        session.put_certificate(SLOT.SIGNATURE, cert, compress=True)
        assert session.get_certificate(SLOT.AUTHENTICATION) == session.get_certificate(
            SLOT.SIGNATURE
        )
        obj1 = session.get_object(OBJECT_ID.from_slot(SLOT.AUTHENTICATION))
        obj2 = session.get_object(OBJECT_ID.from_slot(SLOT.SIGNATURE))
        assert obj1 != obj2
        assert len(obj1) > len(obj2)


class TestManagementKeyReadOnly:
    """
    Tests after which the management key is always the default management
    key. Placing compatible tests here reduces the amount of slow reset
    calls needed.
    """

    def test_authenticate_twice_does_not_throw(self, session, keys):
        session.authenticate(keys.mgmt)
        session.authenticate(keys.mgmt)

    def test_reset_resets_has_stored_key_flag(self, session, keys, scp):
        pivman = get_pivman_data(session)
        assert not pivman.has_stored_key

        session.verify_pin(keys.pin)
        session.authenticate(keys.mgmt)
        pivman_set_mgm_key(
            session,
            NON_DEFAULT_MANAGEMENT_KEY,
            session.management_key_type,
            store_on_device=True,
        )

        pivman = get_pivman_data(session)
        assert pivman.has_stored_key

        reset_state(session, scp)
        session.reset()

        pivman = get_pivman_data(session)
        assert not pivman.has_stored_key

    # Should this really fail?
    def disabled_test_reset_while_verified_throws_nice_ValueError(self, session, keys):
        session.verify_pin(keys.pin)
        with pytest.raises(ValueError) as cm:
            session.reset()
        assert "Cannot read remaining tries from status word: 9000" in str(cm.exception)

    def test_set_mgm_key_does_not_change_key_if_not_authenticated(self, session, keys):
        with pytest.raises(ApduError):
            session.set_management_key(
                session.management_key_type, NON_DEFAULT_MANAGEMENT_KEY
            )
        assert_mgm_key_is(session, keys.mgmt)

    @condition.min_version(3, 5)
    def test_set_stored_mgm_key_does_not_destroy_key_if_pin_not_verified(
        self, session, keys
    ):
        session.authenticate(keys.mgmt)
        with pytest.raises(ApduError):
            pivman_set_mgm_key(
                session,
                NON_DEFAULT_MANAGEMENT_KEY,
                session.management_key_type,
                store_on_device=True,
            )

        assert_mgm_key_is(session, keys.mgmt)


class TestManagementKeyReadWrite:
    """
    Tests after which the management key may not be the default management
    key.
    """

    def test_set_mgm_key_changes_mgm_key(self, session, keys):
        session.authenticate(keys.mgmt)
        session.set_management_key(
            session.management_key_type, NON_DEFAULT_MANAGEMENT_KEY
        )

        assert_mgm_key_is_not(session, keys.mgmt)
        assert_mgm_key_is(session, NON_DEFAULT_MANAGEMENT_KEY)

    def test_set_stored_mgm_key_succeeds_if_pin_is_verified(self, session, keys):
        session.verify_pin(keys.pin)
        session.authenticate(keys.mgmt)
        pivman_set_mgm_key(
            session,
            NON_DEFAULT_MANAGEMENT_KEY,
            session.management_key_type,
            store_on_device=True,
        )

        assert_mgm_key_is_not(session, keys.mgmt)
        assert_mgm_key_is(session, NON_DEFAULT_MANAGEMENT_KEY)

        pivman_prot = get_pivman_protected_data(session)
        assert pivman_prot.key == NON_DEFAULT_MANAGEMENT_KEY

        pivman_prot = get_pivman_protected_data(session)
        assert_mgm_key_is(session, pivman_prot.key)


def sign(session, slot, key_type, message):
    return session.sign(slot, key_type, message, hashes.SHA256(), padding.PKCS1v15())


class TestOperations:
    @condition.min_version(4)
    def test_sign_with_pin_policy_always_requires_pin_every_time(
        self, session, keys, scp
    ):
        generate_key(session, scp, keys, pin_policy=PIN_POLICY.ALWAYS)

        with pytest.raises(ApduError):
            sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")

        session.verify_pin(keys.pin)
        sig = sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")
        assert sig

        with pytest.raises(ApduError):
            sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")

        session.verify_pin(keys.pin)
        sig = sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")
        assert sig

    @condition.yk4_fips(False)
    @condition.check(lambda info: CAPABILITY.PIV not in info.fips_capable)
    @condition.min_version(4)
    def test_sign_with_pin_policy_never_does_not_require_pin(self, session, keys, scp):
        generate_key(session, scp, keys, pin_policy=PIN_POLICY.NEVER)
        sig = sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")
        assert sig

    @condition.yk4_fips(True)
    def test_pin_policy_never_blocked_on_fips(self, session, keys, scp):
        with pytest.raises(NotSupportedError):
            generate_key(session, scp, keys, pin_policy=PIN_POLICY.NEVER)

    @condition.min_version(4)
    def test_sign_with_pin_policy_once_requires_pin_once_per_session(
        self, session, keys, scp
    ):
        generate_key(session, scp, keys, pin_policy=PIN_POLICY.ONCE)

        with pytest.raises(ApduError):
            sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")

        session.verify_pin(keys.pin)
        sig = sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")
        assert sig

        sig = sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")
        assert sig

        reset_state(session, scp)

        with pytest.raises(ApduError):
            sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")

        session.verify_pin(keys.pin)
        sig = sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")
        assert sig

        sig = sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, b"foo")
        assert sig

    def test_signature_can_be_verified_by_public_key(self, session, keys, scp):
        public_key = generate_key(session, scp, keys)

        signed_data = bytes(random.randint(0, 255) for i in range(32))

        session.verify_pin(keys.pin)
        sig = sign(session, SLOT.AUTHENTICATION, KEY_TYPE.ECCP256, signed_data)
        assert sig

        public_key.verify(sig, signed_data, ec.ECDSA(hashes.SHA256()))


def block_pin(session):
    while session.get_pin_attempts() > 0:
        try:
            session.verify_pin(NON_DEFAULT_PIN)
        except Exception:
            pass


class TestUnblockPin:
    @pytest.fixture(autouse=True)
    def preconditions(self, bio_metadata):
        if bio_metadata:
            pytest.skip("PUK not supported on this YubiKey")

    def test_unblock_pin_requires_no_previous_authentication(self, session, keys):
        session.unblock_pin(keys.puk, NON_DEFAULT_PIN)

    def test_unblock_pin_with_wrong_puk_throws_InvalidPinError(self, session):
        with pytest.raises(InvalidPinError):
            session.unblock_pin(NON_DEFAULT_PUK, NON_DEFAULT_PIN)

    def test_unblock_pin_resets_pin_and_retries(self, session, keys):
        block_pin(session)

        with pytest.raises(InvalidPinError):
            session.verify_pin(keys.pin)

        session.unblock_pin(keys.puk, NON_DEFAULT_PIN)

        assert session.get_pin_attempts() == 3
        session.verify_pin(NON_DEFAULT_PIN)

    def test_set_pin_retries_requires_pin_and_mgm_key(
        self, session, version, default_keys, scp
    ):
        keys = default_keys

        # Fails with no authentication
        with pytest.raises(ApduError):
            session.set_pin_attempts(4, 4)

        # Fails with only PIN
        session.verify_pin(keys.pin)
        with pytest.raises(ApduError):
            session.set_pin_attempts(4, 4)

        reset_state(session, scp)

        session.authenticate(keys.mgmt)
        # Fails with only management key (requirement added in 0.1.3)
        if session.version >= (0, 1, 3):
            with pytest.raises(ApduError):
                session.set_pin_attempts(4, 4)

        # Succeeds with both PIN and management key
        session.verify_pin(keys.pin)
        session.set_pin_attempts(4, 4)

    def test_set_pin_retries_sets_pin_and_puk_tries(self, session, default_keys, scp):
        keys = default_keys
        pin_tries = 9
        puk_tries = 7

        session.verify_pin(keys.pin)
        session.authenticate(keys.mgmt)
        session.set_pin_attempts(pin_tries, puk_tries)

        reset_state(session, scp)

        assert session.get_pin_attempts() == pin_tries

        with pytest.raises(InvalidPinError) as ctx:
            session.verify_pin(NON_DEFAULT_PIN)
        assert ctx.value.attempts_remaining == pin_tries - 1

        with pytest.raises(InvalidPinError) as ctx:
            session.change_puk(NON_DEFAULT_PUK, keys.puk)
        assert ctx.value.attempts_remaining == puk_tries - 1


class TestMetadata:
    @pytest.fixture(autouse=True)
    @condition.min_version(5, 3)
    def preconditions(self):
        pass

    def test_pin_metadata(self, session, bio_metadata):
        data = session.get_pin_metadata()
        assert data.default_value is True
        assert data.total_attempts == 8 if bio_metadata else 3
        assert data.attempts_remaining == data.total_attempts

    def test_management_key_metadata(self, session, info):
        data = session.get_management_key_metadata()
        default_type = data.key_type
        if info.version < (5, 7, 0):
            assert data.key_type == MANAGEMENT_KEY_TYPE.TDES
        else:
            assert data.key_type == MANAGEMENT_KEY_TYPE.AES192
        assert data.default_value is True
        assert data.touch_policy is TOUCH_POLICY.NEVER

        session.authenticate(DEFAULT_MANAGEMENT_KEY)
        session.set_management_key(
            MANAGEMENT_KEY_TYPE.AES192, NON_DEFAULT_MANAGEMENT_KEY
        )
        assert session.management_key_type == MANAGEMENT_KEY_TYPE.AES192

        data = session.get_management_key_metadata()
        assert data.key_type == MANAGEMENT_KEY_TYPE.AES192
        assert data.default_value is False
        assert data.touch_policy is TOUCH_POLICY.NEVER

        session.set_management_key(default_type, DEFAULT_MANAGEMENT_KEY)
        data = session.get_management_key_metadata()
        assert data.default_value is True

        if CAPABILITY.PIV not in info.fips_capable:
            session.set_management_key(
                MANAGEMENT_KEY_TYPE.TDES, NON_DEFAULT_MANAGEMENT_KEY
            )
            data = session.get_management_key_metadata()
            assert data.default_value is False

        session.reset()
        assert session.management_key_type == default_type

    @pytest.mark.parametrize("key_type", list(KEY_TYPE))
    def test_slot_metadata_generate(self, session, info, keys, key_type, scp):
        skip_unsupported_key_type(key_type, info)

        slot = SLOT.SIGNATURE
        key = generate_key(session, scp, keys, slot, key_type)
        data = session.get_slot_metadata(slot)

        assert data.key_type == key_type
        assert data.pin_policy == PIN_POLICY.ALWAYS
        assert data.touch_policy == TOUCH_POLICY.NEVER
        assert data.generated is True
        assert data.public_key.public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        ) == key.public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    @pytest.mark.parametrize(
        "key",
        [
            rsa.generate_private_key(65537, 1024, default_backend()),
            rsa.generate_private_key(65537, 2048, default_backend()),
            ec.generate_private_key(ec.SECP256R1(), default_backend()),
            ec.generate_private_key(ec.SECP384R1(), default_backend()),
        ],
    )
    @pytest.mark.parametrize(
        "slot, pin_policy",
        [
            (SLOT.AUTHENTICATION, PIN_POLICY.ONCE),
            (SLOT.SIGNATURE, PIN_POLICY.ALWAYS),
            (SLOT.KEY_MANAGEMENT, PIN_POLICY.ONCE),
            (SLOT.CARD_AUTH, PIN_POLICY.NEVER),
        ],
    )
    def test_slot_metadata_put(self, session, info, keys, key, slot, pin_policy):
        key_type = KEY_TYPE.from_public_key(key.public_key())
        skip_unsupported_key_type(key_type, info, pin_policy)
        session.authenticate(keys.mgmt)
        session.put_key(slot, key)
        data = session.get_slot_metadata(slot)

        assert data.key_type == KEY_TYPE.from_public_key(key.public_key())
        assert data.pin_policy == pin_policy
        assert data.touch_policy == TOUCH_POLICY.NEVER
        assert data.generated is False
        assert data.public_key.public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        ) == key.public_key().public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )


class TestMoveAndDelete:
    @pytest.fixture(autouse=True)
    @condition.min_version(5, 7)
    def preconditions(self):
        pass

    def test_move_key(self, session, keys):
        key = ec.generate_private_key(ec.SECP256R1(), default_backend())
        session.authenticate(keys.mgmt)
        session.put_key(SLOT.AUTHENTICATION, key)
        data_a = session.get_slot_metadata(SLOT.AUTHENTICATION)

        session.move_key(SLOT.AUTHENTICATION, SLOT.SIGNATURE)
        data_s = session.get_slot_metadata(SLOT.SIGNATURE)

        assert data_a == data_s
        with pytest.raises(ApduError):
            session.get_slot_metadata(SLOT.AUTHENTICATION)

    def test_delete_key(self, session, keys):
        key = ec.generate_private_key(ec.SECP256R1(), default_backend())
        session.authenticate(keys.mgmt)
        session.put_key(SLOT.AUTHENTICATION, key)
        session.get_slot_metadata(SLOT.AUTHENTICATION)

        session.delete_key(SLOT.AUTHENTICATION)
        with pytest.raises(ApduError):
            session.get_slot_metadata(SLOT.AUTHENTICATION)


class TestPinComplexity:
    @pytest.fixture(autouse=True)
    def preconditions(self, info):
        if not info.pin_complexity:
            pytest.skip("Requires YubiKey with PIN complexity enabled")

    @pytest.mark.parametrize("pin", ("111111", "22222222", "333333", "4444444"))
    def test_repeated_pins(self, session, keys, pin):
        with pytest.raises(ApduError):
            session.change_pin(keys.pin, pin)

    @condition.check(lambda info: info.version_qualifier.type == RELEASE_TYPE.FINAL)
    @pytest.mark.parametrize("pin", ("abc123", "password", "123123"))
    def test_invalid_pins_standard(self, session, keys, pin):
        with pytest.raises(ApduError):
            session.change_pin(keys.pin, pin)

    @condition.check(lambda info: info.version_qualifier.type != RELEASE_TYPE.FINAL)
    @pytest.mark.parametrize("pin", ("1234", "567890", "11223344"))
    def test_invalid_pins_prerelease(self, session, keys, pin):
        with pytest.raises(ApduError):
            session.change_pin(keys.pin, pin)


@pytest.fixture
def bio_metadata(session):
    try:
        return session.get_bio_metadata()
    except NotSupportedError:
        return None


class TestBioMpe:
    @pytest.fixture(autouse=True)
    def preconditions(self, bio_metadata):
        if not bio_metadata:
            pytest.skip("Requires YubiKey Bio with PIV")

    def test_verify_uv_without_fingerprints(self, session, bio_metadata):
        with pytest.raises(InvalidPinError):
            session.verify_uv(check_only=True)



from __future__ import annotations

import logging
from enum import IntEnum, unique
from typing import Mapping, Sequence, cast

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from .core import BadResponseError, Tlv, Version, int2bytes
from .core.smartcard import (
    AID,
    SW,
    ApduError,
    ScpProcessor,
    SmartCardConnection,
    SmartCardProtocol,
)
from .core.smartcard.scp import (
    INS_EXTERNAL_AUTHENTICATE,
    INS_INITIALIZE_UPDATE,
    INS_INTERNAL_AUTHENTICATE,
    INS_PERFORM_SECURITY_OPERATION,
    KeyRef,
    ScpKeyParams,
    ScpKid,
    StaticKeys,
)

logger = logging.getLogger(__name__)


INS_GET_DATA = 0xCA
INS_PUT_KEY = 0xD8
INS_STORE_DATA = 0xE2
INS_DELETE = 0xE4
INS_GENERATE_KEY = 0xF1

TAG_KEY_INFORMATION = 0xE0
TAG_CARD_RECOGNITION_DATA = 0x66
TAG_CA_KLOC_IDENTIFIERS = 0xFF33
TAG_CA_KLCC_IDENTIFIERS = 0xFF34
TAG_CERTIFICATE_STORE = 0xBF21


@unique
class KeyType(IntEnum):
    AES = 0x88
    ECC_PUBLIC_KEY = 0xB0
    ECC_PRIVATE_KEY = 0xB1
    ECC_KEY_PARAMS = 0xF0


_DEFAULT_KCV_IV = b"\1" * 16


@unique
class Curve(IntEnum):
    SECP256R1 = 0x00
    SECP384R1 = 0x01
    SECP521R1 = 0x02
    BrainpoolP256R1 = 0x03
    BrainpoolP384R1 = 0x05
    BrainpoolP512R1 = 0x07

    @classmethod
    def _from_key(
        cls, key: ec.EllipticCurvePrivateKey | ec.EllipticCurvePublicKey
    ) -> Curve:
        name = key.curve.name.lower()
        for curve in cls:
            if curve.name.lower() == name:
                return curve
        raise ValueError("Unsupported private key")

    @property
    def _curve(self) -> ec.EllipticCurve:
        return getattr(ec, self.name)()


def _int2asn1(value: int) -> bytes:
    bs = int2bytes(value)
    if bs[0] & 0x80:
        bs = b"\x00" + bs
    return Tlv(0x93, bs)


def _encrypt_cbc(key: bytes, data: bytes, iv: bytes = b"\0" * 16) -> bytes:
    encryptor = Cipher(
        algorithms.AES(key),
        modes.CBC(iv),
        backend=default_backend(),
    ).encryptor()
    return encryptor.update(data) + encryptor.finalize()


class SecurityDomainSession:
    """A session for managing SCP keys"""

    def __init__(self, connection: SmartCardConnection):
        self.protocol = SmartCardProtocol(connection)
        self.protocol.select(AID.SECURE_DOMAIN)
        # We don't know the exact version, but this is the minimum that supports SCP
        self.protocol.configure(Version(5, 3, 0))
        logger.debug("SecurityDomain session initialized")

    def authenticate(self, key_params: ScpKeyParams) -> None:
        """Initialize SCP and authenticate the session.

        SCP11b does not authenticate the OCE, and will not allow the usage of commands
        which require authentication of the OCE.
        """
        self.protocol.init_scp(key_params)

    def get_data(self, tag: int, data: bytes = b"") -> bytes:
        """Read data from the security domain."""
        return self.protocol.send_apdu(0, INS_GET_DATA, tag >> 8, tag & 0xFF, data)

    def get_key_information(self) -> Mapping[KeyRef, Mapping[int, int]]:
        """Get information about the currently loaded keys."""
        # 11.3.3.1.1 Key Information Template ('E0')
        keys = {}
        for d in Tlv.parse_list(self.get_data(TAG_KEY_INFORMATION)):
            data = Tlv.unpack(0xC0, d)
            keys[KeyRef(data[:2])] = dict(zip(data[2::2], data[3::2]))
        return keys

    def get_card_recognition_data(self) -> bytes:
        """Get information about the card."""
        # 7.4.1.3 Card Recognition Data
        return Tlv.unpack(0x73, self.get_data(TAG_CARD_RECOGNITION_DATA))

    def get_supported_ca_identifiers(
        self, kloc: bool = False, klcc: bool = False
    ) -> Mapping[KeyRef, bytes]:
        """Get a list of the CA issuer Subject Key Identifiers for keys.

        Setting one of kloc or klcc to True will cause only those CAs to be returned.
        By default, this will get both KLOC and KLCC CAs.

        :param kloc: Get KLOC CAs.
        :param klcc: Get KLCC CAs.
        """
        if not kloc and not klcc:
            kloc = klcc = True
        logger.debug(f"Getting CA identifiers KLOC={kloc}, KLCC={klcc}")
        data = b""
        # Combine CA list for KLCC and KLOC
        for fetch, tag in (
            (kloc, TAG_CA_KLOC_IDENTIFIERS),
            (klcc, TAG_CA_KLCC_IDENTIFIERS),
        ):
            if fetch:
                try:
                    data += self.get_data(tag)
                except ApduError as e:
                    if e.sw != SW.REFERENCE_DATA_NOT_FOUND:
                        raise
        tlvs = Tlv.parse_list(data)
        return {
            KeyRef(tlvs[i + 1].value): tlvs[i].value for i in range(0, len(tlvs), 2)
        }

    def get_certificate_bundle(self, key: KeyRef) -> Sequence[x509.Certificate]:
        """Get the certificates associated with the given SCP11 private key.

        Certificates are returned leaf-last.
        """
        logger.debug(f"Getting certificate bundle for {key}")
        try:
            return [
                x509.load_der_x509_certificate(cert)
                for cert in Tlv.parse_list(
                    self.get_data(TAG_CERTIFICATE_STORE, Tlv(0xA6, Tlv(0x83, key)))
                )
            ]
        except ApduError as e:
            if e.sw == SW.REFERENCE_DATA_NOT_FOUND:
                return []
            raise

    def reset(self) -> None:
        """Perform a factory reset of the Security Domain.

        This will remove all keys and associated data, as well as restore the default
        SCP03 static keys, and generate a new (attestable) SCP11b key.
        """
        logger.debug("Resetting all SCP keys")
        # Reset is done by blocking all available keys
        data = b"\0" * 8
        for key in self.get_key_information().keys():
            if key.kid == 0x01:
                # SCP03 uses KID=0, we use KVN=0 to allow deleting the default keys
                # which have an invalid KVN (0xff).
                key = KeyRef(0, 0)
                ins = INS_INITIALIZE_UPDATE
            elif key.kid in (0x02, 0x03):
                continue  # Skip these, will be deleted by 0x01
            elif key.kid in (0x11, 0x15):
                ins = INS_EXTERNAL_AUTHENTICATE
            elif key.kid == 0x13:
                ins = INS_INTERNAL_AUTHENTICATE
            else:  # 0x10, 0x20-0x2F
                ins = INS_PERFORM_SECURITY_OPERATION

            for _ in range(65):
                try:
                    self.protocol.send_apdu(0x80, ins, key.kvn, key.kid, data)
                except ApduError as e:
                    if e.sw in (
                        SW.AUTH_METHOD_BLOCKED,
                        SW.SECURITY_CONDITION_NOT_SATISFIED,
                    ):
                        break
                    elif e.sw == SW.INCORRECT_PARAMETERS:
                        continue
                    raise
        logger.info("SCP keys reset")

    def store_data(self, data: bytes) -> None:
        """Stores data in the security domain.

        Requires OCE verification.
        """
        self.protocol.send_apdu(0, INS_STORE_DATA, 0x90, 0, data)

    def store_certificate_bundle(
        self, key: KeyRef, certificates: Sequence[x509.Certificate]
    ) -> None:
        """Store the certificate chain for the given key.

        Requires OCE verification.

        Certificates should be in order, with the leaf certificate last.
        """
        logger.debug(f"Storing certificate bundle for {key}")
        self.store_data(
            Tlv(0xA6, Tlv(0x83, key))
            + Tlv(
                TAG_CERTIFICATE_STORE,
                b"".join(
                    c.public_bytes(serialization.Encoding.DER) for c in certificates
                ),
            )
        )
        logger.info("Certificate bundle stored")

    def store_allowlist(self, key: KeyRef, serials: Sequence[int]) -> None:
        """Store which certificate serial numbers that can be used for a given key.

        Requires OCE verification.

        If no allowlist is stored, any certificate signed by the CA can be used.
        """
        logger.debug(f"Storing serial allowlist for {key}")
        self.store_data(
            Tlv(0xA6, Tlv(0x83, key))
            + Tlv(0x70, b"".join(_int2asn1(s) for s in serials))
        )
        logger.info("Serial allowlist stored")

    def store_ca_issuer(self, key: KeyRef, ski: bytes) -> None:
        """Store the SKI (Subject Key Identifier) for the CA of a given key.

        Requires OCE verification.
        """
        logger.debug(f"Storing CA issuer SKI for {key}: {ski.hex()}")
        klcc = key.kid in (ScpKid.SCP11a, ScpKid.SCP11b, ScpKid.SCP11c)
        self.store_data(
            Tlv(
                0xA6,
                Tlv(0x80, b"\1" if klcc else b"\0") + Tlv(0x42, ski) + Tlv(0x83, key),
            )
        )
        logger.info("CA issuer SKI stored")

    def delete_key(self, kid: int = 0, kvn: int = 0, delete_last: bool = False) -> None:
        """Delete one (or more) keys.

        Requires OCE verification.

        All keys matching the given KID and/or KVN will be deleted.
        To delete the final key you must set delete_last = True.
        """
        if not kid and not kvn:
            raise ValueError("Must specify at least one of kid, kvn.")

        if kid in (1, 2, 3):
            # SCP03 keys can only be deleted by KVN
            if kvn:
                kid = 0
            else:
                raise ValueError("SCP03 keys can only be deleted by KVN")
        logger.debug(f"Deleting keys with KID={kid or 'ANY'}, KVN={kvn or 'ANY'}")
        data = b""
        if kid:
            data += Tlv(0xD0, bytes([kid]))
        if kvn:
            data += Tlv(0xD2, bytes([kvn]))
        self.protocol.send_apdu(0x80, INS_DELETE, 0, int(delete_last), data)
        logger.info("Keys deleted")

    def generate_ec_key(
        self, key: KeyRef, curve: Curve = Curve.SECP256R1, replace_kvn: int = 0
    ) -> ec.EllipticCurvePublicKey:
        """Generate a new SCP11 key.

        Requires OCE verification.

        Use replace_kvn to replace an existing key.
        """
        logger.debug(
            f"Generating new key for {key}"
            + (f", replacing KVN={replace_kvn}" if replace_kvn else "")
        )
        data = bytes([key.kvn]) + Tlv(KeyType.ECC_KEY_PARAMS, bytes([curve]))
        resp = self.protocol.send_apdu(
            0x80, INS_GENERATE_KEY, replace_kvn, key.kid, data
        )
        encoded_point = Tlv.unpack(KeyType.ECC_PUBLIC_KEY, resp)
        logger.info("New key generated")
        return ec.EllipticCurvePublicKey.from_encoded_point(curve._curve, encoded_point)

    def put_key(
        self,
        key: KeyRef,
        sk: StaticKeys | ec.EllipticCurvePrivateKey | ec.EllipticCurvePublicKey,
        replace_kvn: int = 0,
    ) -> None:
        """Import an SCP key.

        Requires OCE verification.

        The value of the sk argument should match the SCP type as defined by the KID.
        Use replace_kvn to replace an existing key.
        """
        logger.debug(f"Importing key into {key} of type {type(sk)}")
        processor = self.protocol._processor
        if not isinstance(processor, ScpProcessor):
            raise ValueError("Must be authenticated!")

        data = bytes([key.kvn])
        expected = data
        dek = processor.state._keys.key_dek
        p2 = key.kid
        if isinstance(sk, StaticKeys):
            if not dek:
                raise ValueError("No session DEK key available")
            if not sk.key_dek:
                raise ValueError("New DEK must be set in static keys")
            p2 |= 0x80
            for k in cast(Sequence[bytes], sk):
                kcv = _encrypt_cbc(k, _DEFAULT_KCV_IV)[:3]
                data += Tlv(KeyType.AES, _encrypt_cbc(dek, k)) + bytes([len(kcv)]) + kcv
                expected += kcv
        else:
            if isinstance(sk, ec.EllipticCurvePrivateKey):
                if not dek:
                    raise ValueError("No session DEK key available")
                n = (sk.key_size + 7) // 8
                s = int2bytes(sk.private_numbers().private_value, n)
                data += Tlv(KeyType.ECC_PRIVATE_KEY, _encrypt_cbc(dek, s))
            elif isinstance(sk, ec.EllipticCurvePublicKey):
                data += Tlv(
                    KeyType.ECC_PUBLIC_KEY,
                    sk.public_bytes(
                        serialization.Encoding.X962,
                        serialization.PublicFormat.UncompressedPoint,
                    ),
                )
            else:
                raise TypeError("Unsupported key type")
            data += Tlv(KeyType.ECC_KEY_PARAMS, bytes([Curve._from_key(sk)])) + b"\0"

        resp = self.protocol.send_apdu(0x80, INS_PUT_KEY, replace_kvn, p2, data)
        if resp != expected:
            raise BadResponseError("Incorrect key check value")
        logger.info("Key imported")



# Copyright (c) 2015 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import ctypes
import logging
import re
import sys
import time
from dataclasses import replace

import click
import click.shell_completion
from cryptography.exceptions import InvalidSignature

from yubikit.core import ApplicationNotAvailableError, _override_version
from yubikit.core.fido import FidoConnection
from yubikit.core.otp import OtpConnection
from yubikit.core.smartcard import SmartCardConnection
from yubikit.core.smartcard.scp import (
    KeyRef,
    Scp03KeyParams,
    ScpKid,
    StaticKeys,
)
from yubikit.logging import LOG_LEVEL
from yubikit.management import RELEASE_TYPE
from yubikit.support import get_name, read_info

from .. import __version__
from ..device import list_all_devices as _list_all_devices
from ..device import scan_devices
from ..diagnostics import get_diagnostics, sys_info
from ..logging import init_logging
from ..pcsc import list_devices as list_ccid
from ..pcsc import list_readers
from ..settings import AppData
from ..util import (
    InvalidPasswordError,
    get_windows_version,
    is_nfc_restricted,
    parse_certificates,
    parse_private_key,
)
from .apdu import apdu
from .config import config
from .fido import fido
from .hsmauth import hsmauth
from .info import info
from .oath import oath
from .openpgp import openpgp
from .otp import otp
from .piv import piv
from .script import run_script
from .securitydomain import ScpKidParamType, click_parse_scp_ref, securitydomain
from .util import (
    CliFail,
    EnumChoice,
    HexIntParamType,
    YkmanContextObject,
    click_group,
    click_prompt,
    find_scp11_params,
    organize_scp11_certificates,
    pretty_print,
)

logger = logging.getLogger(__name__)


CLICK_CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"], max_content_width=999)


WIN_CTAP_RESTRICTED = (
    sys.platform == "win32"
    and not bool(ctypes.windll.shell32.IsUserAnAdmin())
    and get_windows_version() >= (10, 0, 18362)
)


def _scan_changes(state, attempts=10):
    for _ in range(attempts):
        time.sleep(0.25)
        devices, new_state = scan_devices()
        if new_state != state:
            return devices, new_state
    raise TimeoutError("Timed out waiting for state change")


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(f"YubiKey Manager (ykman) version: {__version__}")
    ctx.exit()


def print_diagnostics(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo("\n".join(pretty_print(get_diagnostics())))
    ctx.exit()


def require_reader(connection_types, reader):
    if SmartCardConnection in connection_types or FidoConnection in connection_types:
        readers = list_ccid(reader)
        if len(readers) == 1:
            dev = readers[0]
            nfc_restricted = False
            try:
                with dev.open_connection(SmartCardConnection) as conn:
                    try:
                        info = read_info(conn, dev.pid)
                        return dev, info
                    except ValueError:
                        nfc_restricted = is_nfc_restricted(conn)
                        raise  # Re-raise to be handled in block below
            except Exception:
                if nfc_restricted:
                    raise CliFail(
                        "YubiKey is in NFC restricted mode "
                        "(see: https://www.yubico.com/getting-started/)."
                    )
                raise CliFail("Failed to connect to YubiKey.")
        elif len(readers) > 1:
            raise CliFail("Multiple external readers match name.")
        else:
            raise CliFail("No YubiKey found on external reader.")
    else:
        raise CliFail("Not a CCID command.")


def list_all_devices(*args, **kwargs):
    devices = _list_all_devices(*args, **kwargs)
    with_serial = [(dev, dev_info) for (dev, dev_info) in devices if dev_info.serial]
    if with_serial:
        history = AppData("history")
        cache = history.setdefault("devices", {})
        for dev, dev_info in with_serial:
            if dev_info.serial:
                k = str(dev_info.serial)
                cache[k] = cache.pop(k, None) or _describe_device(dev, dev_info, False)
        # 5, chosen by fair dice roll
        [cache.pop(k) for k in list(cache.keys())[: -max(5, len(with_serial))]]
        history.write()
    return devices


def require_device(connection_types, serial=None):
    # Find all connected devices
    devices, state = scan_devices()
    n_devs = sum(devices.values())
    if serial is None:
        if n_devs == 0:  # The device might not yet be ready, wait a bit
            try:
                devices, state = _scan_changes(state)
                n_devs = sum(devices.values())
            except TimeoutError:
                raise CliFail("No YubiKey detected!")
        if n_devs > 1:
            devices_list = list_all_devices()  # Update device cache
            click.echo("Multiple YubiKeys detected:", err=True)
            for dev, dev_info in devices_list:
                click.echo(
                    "- "
                    + _describe_device(dev, dev_info)
                    + (f" Serial: {dev_info.serial}" if dev_info.serial else ""),
                    err=True,
                )
            raise CliFail("Use --device SERIAL to specify which one to use.")

        # Only one connected device, check if any needed interfaces are available
        pid = next(iter(devices.keys()))
        supported = [c for c in connection_types if pid.supports_connection(c)]
        if WIN_CTAP_RESTRICTED and supported == [FidoConnection]:
            # FIDO-only command on Windows without Admin won't work
            raise CliFail("FIDO access on Windows requires running as Administrator.")
        if not supported:
            interfaces = [c.usb_interface for c in connection_types]
            req = ", ".join(t.name or str(t) for t in interfaces)
            raise CliFail(
                f"Command requires one of the following USB interfaces "
                f"to be enabled: '{req}'.\n\n"
                "Use 'ykman config usb' to set the enabled USB interfaces."
            )

        devs = list_all_devices(supported)
        if len(devs) != 1:
            raise CliFail("Failed to connect to YubiKey.")
        return devs[0]
    else:
        for retry in (
            True,
            False,
        ):  # If no match initially, wait a bit for state change.
            devs = list_all_devices(connection_types)
            for dev, dev_info in devs:
                if dev_info.serial == serial:
                    return dev, dev_info
            try:
                if retry:
                    _, state = _scan_changes(state)
            except TimeoutError:
                break

        raise CliFail(
            f"Failed connecting to a YubiKey with serial: {serial}.\n"
            "Make sure the application has the required permissions.",
        )


@click_group(context_settings=CLICK_CONTEXT_SETTINGS)
@click.option(
    "-d",
    "--device",
    type=int,
    metavar="SERIAL",
    help="specify which YubiKey to interact with by serial number",
    shell_complete=lambda ctx, param, incomplete: [
        click.shell_completion.CompletionItem(
            serial,
            help=description,
        )
        for serial, description in AppData("history").get("devices", {}).items()
        if serial.startswith(incomplete)
    ],
)
@click.option(
    "-r",
    "--reader",
    help="specify a YubiKey by smart card reader name "
    "(can't be used with --device or list)",
    metavar="NAME",
    default=None,
    shell_complete=lambda ctx, param, incomplete: [
        f'"{reader.name}"' for reader in list_readers()
    ],
)
@click.option(
    "-t",
    "--scp-ca",
    type=click.File("rb"),
    help="specify the CA to use to verify the SCP11 card key (CA-KLCC)",
)
@click.option(
    "-c",
    "--scp-sd",
    metavar="KID KVN",
    type=(ScpKidParamType(), HexIntParamType()),
    default=(0, 0),
    callback=click_parse_scp_ref,
    hidden="--full-help" not in sys.argv,
    help="specify which key the YubiKey is using to authenticate",
)
@click.option(
    "-o",
    "--scp-oce",
    metavar="KID KVN",
    type=HexIntParamType(),
    nargs=2,
    default=(0, 0),
    hidden="--full-help" not in sys.argv,
    help="specify which key the OCE is using to authenticate",
)
@click.option(
    "-s",
    "--scp",
    "scp_cred",
    metavar="CRED",
    multiple=True,
    help="specify private key and certificate chain for secure messaging, "
    "can be used multiple times to provide key and certificates in multiple "
    "files (private key, certificates in leaf-last order), OR SCP03 keys in hex "
    "separated by colon (:) K-ENC:K-MAC[:K-DEK]",
)
@click.option(
    "-p",
    "--scp-password",
    "scp_cred_password",
    metavar="PASSWORD",
    help="specify a password required to access the --scp file, if needed",
)
@click.option(
    "-l",
    "--log-level",
    default=None,
    type=EnumChoice(LOG_LEVEL, hidden=[LOG_LEVEL.NOTSET]),
    help="enable logging at given verbosity level",
)
@click.option(
    "--log-file",
    default=None,
    type=str,
    metavar="FILE",
    help="write log to FILE instead of printing to stderr (requires --log-level)",
)
@click.option(
    "--diagnose",
    is_flag=True,
    callback=print_diagnostics,
    expose_value=False,
    is_eager=True,
    help="show diagnostics information useful for troubleshooting",
)
@click.option(
    "-v",
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="show version information about the app",
)
@click.option(
    "--full-help",
    is_flag=True,
    expose_value=False,
    help="show --help output, including hidden commands",
)
@click.pass_context
def cli(
    ctx,
    device,
    scp_ca,
    scp_sd,
    scp_oce,
    scp_cred,
    scp_cred_password,
    log_level,
    log_file,
    reader,
):
    """
    Configure your YubiKey via the command line.

    Examples:

    \b
      List connected YubiKeys, only output serial number:
      $ ykman list --serials

    \b
      Show information about YubiKey with serial number 123456:
      $ ykman --device 123456 info
    """
    ctx.obj = YkmanContextObject()

    if log_level:
        init_logging(log_level, log_file=log_file, replace=log_file is None)
        logger.info("\n".join(pretty_print({"System info": sys_info()})))
    elif log_file:
        ctx.fail("--log-file requires specifying --log-level.")

    if reader and device:
        ctx.fail("--reader and --device options can't be combined.")

    use_scp = bool(any(scp_sd) or scp_cred or scp_ca)

    subcmd = next(c for c in COMMANDS if c.name == ctx.invoked_subcommand)
    # Commands that don't directly act on a key
    if subcmd in (list_keys,):
        if device:
            ctx.fail("--device can't be used with this command.")
        if reader:
            ctx.fail("--reader can't be used with this command.")
        if use_scp:
            ctx.fail("SCP can't be used with this command.")
        return

    # FIDO command on Windows without Admin won't work
    if subcmd == fido and WIN_CTAP_RESTRICTED:
        raise CliFail("FIDO access on Windows requires running as Administrator.")

    # Commands which need a YubiKey to act on
    connections = getattr(
        subcmd, "connections", [SmartCardConnection, FidoConnection, OtpConnection]
    )
    if connections:

        def resolve():
            items = getattr(resolve, "items", None)
            if not items:
                # We might be connecting over NFC, and thus may require SCP11
                if reader is not None:
                    items = require_reader(connections, reader)
                else:
                    items = require_device(connections, device)

                if items[1].version_qualifier.type != RELEASE_TYPE.FINAL:
                    # Preview build, override version
                    version_q = items[1].version_qualifier
                    _override_version(version_q.version)
                    logger.info(f"Debug key detected: {version_q}")

                setattr(resolve, "items", items)
            return items

        ctx.obj.add_resolver("device", lambda: resolve()[0])
        ctx.obj.add_resolver("pid", lambda: resolve()[0].pid)
        ctx.obj.add_resolver("info", lambda: resolve()[1])

        if use_scp:
            if SmartCardConnection not in connections:
                raise CliFail("SCP can only be used with CCID commands.")

            scp_kid, scp_kvn = scp_sd
            if scp_kid:
                try:
                    scp_kid = ScpKid(scp_kid)
                except ValueError:
                    raise CliFail(f"Invalid KID for card certificate: {scp_kid}.")

            if scp_ca:
                ca = scp_ca.read()
            else:
                ca = None

            key_fmt = r"[0-9a-fA-F]{32}"
            re_hex_keys = re.compile(rf"^{key_fmt}:{key_fmt}(:{key_fmt})?$")
            if len(scp_cred) == 1 and re_hex_keys.match(scp_cred[0]):
                scp03_keys = StaticKeys(
                    *(bytes.fromhex(k) for k in scp_cred[0].split(":"))
                )
                scp11_creds = None
            else:
                f = click.File("rb")
                scp11_creds = [f.convert(fn, None, ctx).read() for fn in scp_cred]
                scp03_keys = None

            if not scp_kid:
                if scp03_keys:
                    scp_kid = ScpKid.SCP03
                elif not scp11_creds:
                    scp_kid = ScpKid.SCP11b

            if scp03_keys and scp_kid != ScpKid.SCP03:
                raise CliFail("--scp with SCP03 keys can only be used with SCP03.")

            if scp_kid == ScpKid.SCP03:
                if scp_ca:
                    raise CliFail("--scp-ca can only be used with SCP11.")

                def params_f_scp03(_):
                    return Scp03KeyParams(
                        ref=KeyRef(ScpKid.SCP03, scp_kvn),
                        keys=scp03_keys or StaticKeys.default(),
                    )

                params_f = params_f_scp03

            elif scp11_creds:
                # SCP11 a/c
                if scp_kid and scp_kid not in (ScpKid.SCP11a, ScpKid.SCP11c):
                    raise CliFail("--scp with file(s) can only be used with SCP11 a/c.")

                first = scp11_creds.pop(0)
                password = scp_cred_password.encode() if scp_cred_password else None

                while True:
                    try:
                        sk_oce_ecka = parse_private_key(first, password)
                        break
                    except InvalidPasswordError:
                        if scp_cred_password:
                            raise CliFail("Wrong password to decrypt private key.")
                        logger.debug("Error parsing key", exc_info=True)
                        password = click_prompt(
                            "Enter password to decrypt SCP11 key",
                            default="",
                            hide_input=True,
                            show_default=False,
                        ).encode()

                if scp11_creds:
                    certificates = []
                    for c in scp11_creds:
                        certificates.extend(parse_certificates(c, None))
                else:
                    certificates = parse_certificates(first, password)
                    # If the bundle contains the CA we strip it out
                    _, inter, leaf = organize_scp11_certificates(certificates)
                    # Send the KA-KLOC and OCE certificates
                    certificates = list(inter) + [leaf]

                def params_f_scp11ab(conn):
                    if not scp_kid:
                        # TODO: Find key based on CA
                        # Check for SCP11a key, then SCP11c
                        try:
                            params = find_scp11_params(conn, ScpKid.SCP11a, scp_kvn, ca)
                        except (ValueError, InvalidSignature) as e:
                            try:
                                params = find_scp11_params(
                                    conn, ScpKid.SCP11c, scp_kvn, ca
                                )
                            except (ValueError, InvalidSignature):
                                raise e
                    else:
                        params = find_scp11_params(conn, scp_kid, scp_kvn, ca)
                    return replace(
                        params,
                        oce_ref=KeyRef(*scp_oce),
                        sk_oce_ecka=sk_oce_ecka,
                        certificates=certificates,
                    )

                params_f = params_f_scp11ab
            else:
                # SCP11b
                if scp_kid not in (ScpKid.SCP11b, None):
                    raise CliFail(f"{scp_kid.name} requires --scp.")
                if any(scp_oce):
                    raise CliFail("SCP11b cannot be used with --scp-oce.")

                def params_f_scp11b(conn):
                    return find_scp11_params(conn, ScpKid.SCP11b, scp_kvn, ca)

                params_f = params_f_scp11b

            connections = [SmartCardConnection]

            ctx.obj.add_resolver("scp", lambda: params_f)


@cli.command("list")
@click.option(
    "-s",
    "--serials",
    is_flag=True,
    help="output only serial numbers, one per line "
    "(devices without serial will be omitted)",
)
@click.option("-r", "--readers", is_flag=True, help="list available smart card readers")
@click.pass_context
def list_keys(ctx, serials, readers):
    """
    List connected YubiKeys.
    """

    if readers:
        for reader in list_readers():
            click.echo(reader.name)
        ctx.exit()

    # List all attached devices
    pids = set()
    for dev, dev_info in list_all_devices():
        if serials:
            if dev_info.serial:
                click.echo(dev_info.serial)
        else:
            click.echo(
                _describe_device(dev, dev_info)
                + (f" Serial: {dev_info.serial}" if dev_info.serial else "")
            )
        pids.add(dev.pid)

    # Look for FIDO devices that we can't access
    if not serials:
        devs, _ = scan_devices()
        for pid, count in devs.items():
            if pid not in pids:
                for _ in range(count):
                    name = pid.yubikey_type.value
                    mode = pid.name.split("_", 1)[1].replace("_", "+")
                    click.echo(f"{name} [{mode}] <access denied>")


def _describe_device(dev, dev_info, include_mode=True):
    if dev.pid is None:  # Devices from list_all_devices should always have PID.
        raise AssertionError("PID is None")
    name = get_name(dev_info, dev.pid.yubikey_type)
    description = f"{name} ({dev_info.version_name})"
    if include_mode:
        mode = dev.pid.name.split("_", 1)[1].replace("_", "+")
        description += f" [{mode}]"
    return description


COMMANDS = (
    list_keys,
    info,
    otp,
    openpgp,
    oath,
    piv,
    fido,
    config,
    apdu,
    run_script,
    hsmauth,
    securitydomain,
)


for cmd in COMMANDS:
    cli.add_command(cmd)


class _DefaultFormatter(logging.Formatter):
    def __init__(self, show_trace=False):
        self.show_trace = show_trace

    def format(self, record):
        message = f"{record.levelname}: {record.getMessage()}"
        if self.show_trace and record.exc_info:
            message += self.formatException(record.exc_info)
        return message


def main():
    # Set up default logging
    handler = logging.StreamHandler()
    handler.setLevel(logging.WARNING)
    formatter = _DefaultFormatter()
    handler.setFormatter(formatter)
    logging.getLogger().addHandler(handler)

    try:
        # --full-help triggers --help, hidden commands will already have read it by now.
        sys.argv[sys.argv.index("--full-help")] = "--help"
    except ValueError:
        pass  # No --full-help

    try:
        # Allow --device SERIAL to be specified anywhere in the argument list
        index = sys.argv.index("--device")
        if index > 1:
            # Move --device to be the first argument after the command name
            arg = sys.argv.pop(index)
            val = sys.argv.pop(index)
            sys.argv.insert(1, val)
            sys.argv.insert(1, arg)
    except ValueError:
        pass  # No --device

    try:
        cli(obj={}, windows_expand_args=False)
    except Exception as e:
        status = 1
        if isinstance(e, CliFail):
            status = e.status
            msg = e.args[0]
        elif isinstance(e, ApplicationNotAvailableError):
            msg = (
                "The functionality required for this command is not enabled or not "
                "available on this YubiKey."
            )
        elif isinstance(e, ValueError):
            msg = f"{e}"
        else:
            msg = "An unexpected error has occurred"
            formatter.show_trace = True
        logger.exception(msg)
        logging.shutdown()
        sys.exit(status)


if __name__ == "__main__":
    main()



import pytest

from yubikit.core.smartcard import ApplicationNotAvailableError, SmartCardProtocol


def test_select_wrong_app(ccid_connection):
    p = SmartCardProtocol(ccid_connection)
    with pytest.raises(ApplicationNotAvailableError):
        p.select(b"not_a_real_aid")



import logging

from fido2.hid import CtapHidDevice, list_descriptors, open_connection

from yubikit.core import PID, TRANSPORT, Connection
from yubikit.support import read_info

from ..base import REINSERT_STATUS, CancelledException, YkmanDevice
from ..fido import FidoConnection

logger = logging.getLogger(__name__)


class CtapYubiKeyDevice(YkmanDevice):
    """YubiKey FIDO USB HID device"""

    def __init__(self, descriptor):
        super().__init__(TRANSPORT.USB, descriptor.path, PID(descriptor.pid))
        self.descriptor = descriptor

    def supports_connection(self, connection_type):
        return issubclass(CtapHidDevice, connection_type)

    def open_connection(self, connection_type):
        assert isinstance(connection_type, type)  # noqa: S101
        if self.supports_connection(connection_type):
            dev = CtapHidDevice(self.descriptor, open_connection(self.descriptor))
            assert isinstance(dev, Connection)  # noqa: S101
            return dev
        return super().open_connection(connection_type)

    def _do_reinsert(self, reinsert_cb, event):
        removed_state = None
        with self.open_connection(FidoConnection) as conn:
            assert isinstance(conn, Connection)  # noqa: S101
            info = read_info(conn, self.pid)

        reinsert_cb(REINSERT_STATUS.REMOVE)
        logger.debug(f"Waiting for removal of device {self.fingerprint}")
        while not event.wait(0.5):
            keys = list_ctap_devices()
            present = {k.fingerprint for k in keys}
            if removed_state is None:
                if self.fingerprint not in present:
                    logger.debug(f"Removed! {self.fingerprint}")
                    reinsert_cb(REINSERT_STATUS.REINSERT)
                    removed_state = present
            else:
                added = present - removed_state
                if len(added) == 1:
                    dev_fp = next(iter(added))  # Path may have changed
                    logger.debug(f"Inserted! {dev_fp}")
                    key = next(k for k in keys if k.fingerprint == dev_fp)
                    # Update fingerprint and descriptor
                    self._fingerprint = key.fingerprint
                    self.descriptor = key.descriptor
                    with self.open_connection(FidoConnection) as conn:
                        assert isinstance(conn, Connection)  # noqa: S101
                        info2 = read_info(conn, self.pid)
                    if info.serial != info2.serial or info.version != info2.version:
                        raise ValueError(
                            "Reinserted YubiKey does not match the original"
                        )
                    return
                elif len(added) > 1:
                    raise ValueError("Multiple YubiKeys inserted")

        raise CancelledException()


def list_ctap_devices() -> list[CtapYubiKeyDevice]:
    devs = []
    for desc in list_descriptors():
        if desc.vid == 0x1050:
            try:
                devs.append(CtapYubiKeyDevice(desc))
            except ValueError:
                logger.debug(f"Unsupported Yubico device with PID: {desc.pid:02x}")
    return devs



# -*- coding: utf-8 -*-

import re
from base64 import b32encode

import pytest
from ykman.oath import STEAM_CHAR_TABLE
from yubikit.management import CAPABILITY

from .. import condition

URI_HOTP_EXAMPLE = (
    "otpauth://hotp/Example:demo@example.com?"
    "secret=JBSWY3DPK5XXE3DEJ5TE6QKUJA======&issuer=Example&counter=1"
)

URI_TOTP_EXAMPLE = (
    "otpauth://totp/ACME%20Co:john.doe@email.com?"
    "secret=HXDMVJECJJWSRB3HWIZR4IFUGFTMXBOZ&issuer=ACME%20Co"
    "&algorithm=SHA1&digits=6&period=30"
)

URI_TOTP_EXAMPLE_B = (
    "otpauth://totp/ACME%20Co:john.doe.b@email.com?"
    "secret=HXDMVJECJJWSRB3HWIZR4IFUGFTMXBOZ&issuer=ACME%20Co"
    "&algorithm=SHA1&digits=6&period=30"
)

URI_TOTP_EXAMPLE_EXTRA_PARAMETER = (
    "otpauth://totp/ACME%20Co:john.doe.extra@email.com?"
    "secret=HXDMVJECJJWSRB3HWIZR4IFUGFTMXBOZ&issuer=ACME%20Co"
    "&algorithm=SHA1&digits=6&period=30&skid=JKS3424d"
)

PASSWORD = "aaaa"


@pytest.fixture(autouse=True)
@condition.capability(CAPABILITY.OATH)
def preconditions(ykman_cli):
    ykman_cli("oath", "reset", "-f")


@pytest.fixture()
def password(info):
    if CAPABILITY.OATH in info.fips_capable:
        yield PASSWORD
    else:
        yield None


@pytest.fixture()
def accounts_cli(ykman_cli, password):
    if password:
        ykman_cli("oath", "access", "change", "-n", password)

    def fn(*args, **kwargs):
        argv = ["oath", "accounts", *args]
        if password:
            argv += ["-p", password]
        return ykman_cli(*argv, **kwargs)

    yield fn


class TestOATH:
    def test_oath_info(self, ykman_cli):
        output = ykman_cli("oath", "info").output
        assert "version:" in output

    @condition.yk4_fips(False)
    def test_info_does_not_indicate_fips_mode_for_non_fips_key(self, ykman_cli):
        info = ykman_cli("oath", "info").output
        assert "FIPS:" not in info

    def test_oath_add_credential(self, accounts_cli, password):
        accounts_cli("add", "test-name", "abba")
        creds = accounts_cli("list").output
        assert "test-name" in creds

    def test_oath_add_credential_prompt(self, accounts_cli):
        accounts_cli("add", "test-name-2", input="abba")
        creds = accounts_cli("list").output
        assert "test-name-2" in creds

    def test_oath_add_credential_with_space(self, accounts_cli):
        accounts_cli("add", "test-name-space", "ab ba")
        creds = accounts_cli("list").output
        assert "test-name-space" in creds

    def test_oath_hidden_cred(self, accounts_cli):
        accounts_cli("add", "_hidden:name", "abba")
        creds = accounts_cli("code").output
        assert "_hidden:name" not in creds
        creds = accounts_cli("code", "-H").output
        assert "_hidden:name" in creds

    def test_oath_add_uri_hotp(self, accounts_cli):
        accounts_cli("uri", URI_HOTP_EXAMPLE)
        creds = accounts_cli("list").output
        assert "Example:demo" in creds

    def test_oath_add_uri_totp(self, accounts_cli):
        accounts_cli("uri", URI_TOTP_EXAMPLE)
        creds = accounts_cli("list").output
        assert "john.doe" in creds

    def test_oath_add_uri_totp_extra_parameter(self, accounts_cli):
        accounts_cli("uri", URI_TOTP_EXAMPLE_EXTRA_PARAMETER)
        creds = accounts_cli("list").output
        assert "john.doe.extra" in creds

    def test_oath_add_uri_totp_prompt(self, accounts_cli):
        accounts_cli("uri", input=URI_TOTP_EXAMPLE_B)
        creds = accounts_cli("list").output
        assert "john.doe" in creds

    def test_oath_code(self, accounts_cli):
        accounts_cli("add", "test-name2", "abba")
        creds = accounts_cli("code").output
        assert "test-name2" in creds

    def test_oath_code_query_single(self, accounts_cli):
        accounts_cli("add", "query-me", "abba")
        creds = accounts_cli("code", "query-me").output
        assert "query-me" in creds

    def test_oath_code_query_multiple(self, accounts_cli):
        accounts_cli("add", "foo", "abba")
        accounts_cli("add", "query-me", "abba")
        accounts_cli("add", "bar", "abba")
        lines = accounts_cli("code", "query").output.strip().splitlines()
        assert len(lines) == 1
        assert "query-me" in lines[0]

    def test_oath_reset(self, ykman_cli):
        output = ykman_cli("oath", "reset", "-f").output
        assert "Reset complete" in output

    def test_oath_hotp_vectors_6(self, accounts_cli):
        accounts_cli(
            "add",
            "-o",
            "HOTP",
            "testvector",
            b32encode(b"12345678901234567890").decode(),
        )
        for code in ["755224", "287082", "359152", "969429", "338314"]:
            words = accounts_cli("code", "testvector").output.split()
            assert code in words

    def test_oath_hotp_vectors_8(self, accounts_cli):
        accounts_cli(
            "add",
            "-o",
            "HOTP",
            "-d",
            "8",
            "testvector8",
            b32encode(b"12345678901234567890").decode(),
        )
        for code in ["84755224", "94287082", "37359152", "26969429", "40338314"]:
            words = accounts_cli("code", "testvector8").output.split()
            assert code in words

    def test_oath_hotp_code(self, accounts_cli):
        accounts_cli("add", "-o", "HOTP", "hotp-cred", "abba")
        words = accounts_cli("code", "hotp-cred").output.split()
        assert "659165" in words

    def test_oath_hotp_code_single(self, accounts_cli):
        accounts_cli("add", "-o", "HOTP", "hotp-cred", "abba")
        words = accounts_cli("code", "hotp-cred", "--single").output.split()
        assert "659165" in words

    def test_oath_totp_steam_code(self, accounts_cli):
        accounts_cli("add", "Steam:steam-cred", "abba")
        cred = accounts_cli("code", "steam-cred").output.strip()
        code = cred.split()[-1]
        assert 5 == len(code), f"cred wrong length: {code!r}"
        assert all(c in STEAM_CHAR_TABLE for c in code), (
            f"{code!r} contains non-steam characters"
        )

    def test_oath_totp_steam_code_single(self, accounts_cli):
        accounts_cli("add", "Steam:steam-cred", "abba")
        code = accounts_cli("code", "-s", "steam-cred").output.strip()
        assert 5 == len(code), f"cred wrong length: {code!r}"
        assert all(c in STEAM_CHAR_TABLE for c in code), (
            f"{code!r} contains non-steam characters"
        )

    def test_oath_code_output_no_touch(self, accounts_cli):
        accounts_cli("add", "TOTP:normal", "aaaa")
        accounts_cli("add", "Steam:normal", "aaba")
        accounts_cli("add", "-o", "HOTP", "HOTP:normal", "abaa")

        lines = accounts_cli("code").output.strip().splitlines()
        entries = {line.split()[0]: line for line in lines}
        assert "HOTP Account" in entries["HOTP:normal"]

        code = entries["Steam:normal"].split()[-1]
        assert 5 == len(code), f"cred wrong length: {code!r}"
        assert all(c in STEAM_CHAR_TABLE for c in code), (
            f"{code!r} contains non-steam characters"
        )

        code = entries["TOTP:normal"].split()[-1]
        assert 6 == len(code)
        int(code)

    @condition.min_version(4)
    def test_oath_code_output(self, accounts_cli):
        accounts_cli("add", "TOTP:normal", "aaaa")
        accounts_cli("add", "--touch", "TOTP:touch", "aaab")
        accounts_cli("add", "Steam:normal", "aaba")
        accounts_cli("add", "--touch", "Steam:touch", "aabb")
        accounts_cli("add", "-o", "HOTP", "HOTP:normal", "abaa")

        lines = accounts_cli("code").output.strip().splitlines()
        entries = {line.split()[0]: line for line in lines}
        assert "Requires Touch" in entries["TOTP:touch"]
        assert "Requires Touch" in entries["Steam:touch"]
        assert "HOTP Account" in entries["HOTP:normal"]

        code = entries["Steam:normal"].split()[-1]
        assert 5 == len(code), f"cred wrong length: {code!r}"
        assert all(c in STEAM_CHAR_TABLE for c in code), (
            f"{code!r} contains non-steam characters"
        )

        code = entries["TOTP:normal"].split()[-1]
        assert 6 == len(code)
        int(code)

    @condition.min_version(4)
    def test_oath_totp_steam_touch_not_in_code_output(self, accounts_cli):
        accounts_cli("add", "--touch", "Steam:steam-cred", "abba")
        accounts_cli("add", "TOTP:totp-cred", "abba")
        lines = accounts_cli("code").output.strip().splitlines()
        assert "Requires Touch" in lines[0]

    def test_oath_delete(self, accounts_cli):
        accounts_cli("add", "delete-me", "abba")
        accounts_cli("delete", "delete-me", "-f")
        assert "delete-me" not in accounts_cli("list").output

    def test_oath_unicode(self, accounts_cli):
        accounts_cli("add", "😃", "abba")
        accounts_cli("code")
        accounts_cli("list")
        accounts_cli("delete", "😃", "-f")

    @condition.yk4_fips(False)
    @condition.min_version(4, 3, 1)
    def test_oath_sha512(self, accounts_cli):
        accounts_cli("add", "abba", "abba", "--algorithm", "SHA512")
        accounts_cli("delete", "abba", "-f")

    # NEO credential capacity may vary based on configuration
    @condition.min_version(4)
    def test_add_max_creds(self, accounts_cli, version):
        n_creds = 32 if version < (5, 7, 0) else 64
        for i in range(n_creds):
            accounts_cli("add", "test" + str(i), "abba")
            output = accounts_cli("list").output
            lines = output.strip().split("\n")
            assert len(lines) == i + 1

        with pytest.raises(SystemExit):
            accounts_cli("add", "testx", "abba")

    @condition.min_version(5, 3, 1)
    def test_rename(self, accounts_cli):
        accounts_cli("uri", URI_TOTP_EXAMPLE)
        accounts_cli("rename", "john.doe", "Example:user@example.com", "-f")

        creds = accounts_cli("list").output
        assert "john.doe" not in creds
        assert "Example:user@example.com" in creds


class TestOathFips:
    @pytest.fixture(autouse=True)
    @condition.yk4_fips(True)
    def check_fips(self):
        pass

    def test_no_fips_mode_without_password(self, ykman_cli):
        output = ykman_cli("oath", "info").output
        assert re.search(r"FIPS approved:\s+False", output)

    def test_fips_mode_with_password(self, ykman_cli):
        ykman_cli("oath", "access", "change", "-n", PASSWORD)
        output = ykman_cli("oath", "info").output
        assert re.search(r"FIPS approved:\s+True", output)

    def test_sha512_not_supported(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli(
                "oath", "accounts", "add", "abba", "abba", "--algorithm", "SHA512"
            )


class TestPskc:
    def test_add_with_generate(self, accounts_cli):
        """Test adding a credential with --generate flag."""
        result = accounts_cli("add", "generated-cred", "--generate", "-f")
        assert "Generated credential secret (base32):" in result.output

        creds = accounts_cli("list").output
        assert "generated-cred" in creds

    def test_add_with_generate_and_secret_fails(self, accounts_cli):
        """Test that --generate cannot be used with a provided secret."""
        with pytest.raises(SystemExit):
            accounts_cli("add", "fail-cred", "abba", "--generate", "-f")

    def test_export_to_pskc_file(self, accounts_cli, tmp_path):
        """Test exporting a credential to a PSKC file."""
        pskc_file = tmp_path / "export.pskc"
        accounts_cli(
            "add", "export-test", "GEZDGNBVGY3TQOJQ", "-O", str(pskc_file), "-f"
        )
        assert pskc_file.exists()

        # Verify the file contains expected data
        content = pskc_file.read_text()
        assert "pskc" in content.lower()
        assert "export-test" in content

    def test_export_with_issuer(self, accounts_cli, tmp_path):
        """Test exporting a credential with issuer to a PSKC file."""
        pskc_file = tmp_path / "export_issuer.pskc"
        accounts_cli(
            "add",
            "user@example.com",
            "GEZDGNBVGY3TQOJQ",
            "-i",
            "ExampleIssuer",
            "-O",
            str(pskc_file),
            "-f",
        )
        assert pskc_file.exists()

        content = pskc_file.read_text()
        assert "ExampleIssuer" in content

    def test_export_with_passphrase(self, accounts_cli, tmp_path):
        """Test exporting a credential encrypted with a passphrase."""
        pskc_file = tmp_path / "encrypted_pass.pskc"
        accounts_cli(
            "add",
            "passphrase-test",
            "GEZDGNBVGY3TQOJQ",
            "-O",
            str(pskc_file),
            "--pskc-passphrase",
            "testpassword123",
            "-f",
        )
        assert pskc_file.exists()

        # Verify the file is encrypted (contains encryption info)
        content = pskc_file.read_text()
        assert "EncryptedValue" in content or "CipherData" in content

    def test_export_with_preshared_key(self, accounts_cli, tmp_path):
        """Test exporting a credential encrypted with a pre-shared key."""
        pskc_file = tmp_path / "encrypted_key.pskc"
        # 16 bytes = 32 hex characters
        psk = "0123456789abcdef0123456789abcdef"
        accounts_cli(
            "add",
            "psk-test",
            "GEZDGNBVGY3TQOJQ",
            "-O",
            str(pskc_file),
            "--pskc-key",
            psk,
            "-f",
        )
        assert pskc_file.exists()

        content = pskc_file.read_text()
        assert "EncryptedValue" in content or "CipherData" in content

    def test_export_pskc_passphrase_too_short(self, accounts_cli, tmp_path):
        """Test that passphrase must be at least 8 characters."""
        pskc_file = tmp_path / "fail.pskc"
        with pytest.raises(SystemExit):
            accounts_cli(
                "add",
                "fail-test",
                "GEZDGNBVGY3TQOJQ",
                "-O",
                str(pskc_file),
                "--pskc-passphrase",
                "short",
                "-f",
            )

    def test_export_pskc_key_wrong_length(self, accounts_cli, tmp_path):
        """Test that pre-shared key must be 16 bytes."""
        pskc_file = tmp_path / "fail.pskc"
        with pytest.raises(SystemExit):
            accounts_cli(
                "add",
                "fail-test",
                "GEZDGNBVGY3TQOJQ",
                "-O",
                str(pskc_file),
                "--pskc-key",
                "0123456789abcdef",  # Only 8 bytes
                "-f",
            )

    def test_export_pskc_key_and_passphrase_fails(self, accounts_cli, tmp_path):
        """Test that --pskc-key and --pskc-passphrase cannot be used together."""
        pskc_file = tmp_path / "fail.pskc"
        with pytest.raises(SystemExit):
            accounts_cli(
                "add",
                "fail-test",
                "GEZDGNBVGY3TQOJQ",
                "-O",
                str(pskc_file),
                "--pskc-key",
                "0123456789abcdef0123456789abcdef",
                "--pskc-passphrase",
                "testpassword",
                "-f",
            )

    def test_pskc_encryption_without_output_fails(self, accounts_cli):
        """Test that encryption options require --output."""
        with pytest.raises(SystemExit):
            accounts_cli(
                "add",
                "fail-test",
                "GEZDGNBVGY3TQOJQ",
                "--pskc-passphrase",
                "testpassword123",
                "-f",
            )

    def test_generate_and_export(self, accounts_cli, tmp_path):
        """Test generating a secret and exporting to PSKC."""
        pskc_file = tmp_path / "generated_export.pskc"
        accounts_cli("add", "gen-export-test", "--generate", "-O", str(pskc_file), "-f")
        assert pskc_file.exists()

        creds = accounts_cli("list").output
        assert "gen-export-test" in creds

    def test_import_pskc_file(self, accounts_cli, tmp_path):
        """Test importing credentials from a PSKC file."""
        # First export a credential
        pskc_file = tmp_path / "import_test.pskc"
        accounts_cli(
            "add",
            "to-import",
            "GEZDGNBVGY3TQOJQ",
            "-i",
            "TestIssuer",
            "-O",
            str(pskc_file),
            "-f",
        )

        # Delete the credential
        accounts_cli("delete", "to-import", "-f")
        creds = accounts_cli("list").output
        assert "to-import" not in creds

        # Re-import from PSKC
        accounts_cli("import", str(pskc_file), "-f")
        creds = accounts_cli("list").output
        assert "to-import" in creds

    def test_import_pskc_encrypted_with_passphrase(self, accounts_cli, tmp_path):
        """Test importing credentials from a passphrase-encrypted PSKC file."""
        passphrase = "testpassword123"
        pskc_file = tmp_path / "encrypted_import.pskc"

        # Export with encryption
        accounts_cli(
            "add",
            "encrypted-import",
            "GEZDGNBVGY3TQOJQ",
            "-O",
            str(pskc_file),
            "--pskc-passphrase",
            passphrase,
            "-f",
        )

        # Delete the credential
        accounts_cli("delete", "encrypted-import", "-f")

        # Re-import with passphrase
        accounts_cli("import", str(pskc_file), "-f", input=passphrase)
        creds = accounts_cli("list").output
        assert "encrypted-import" in creds

    def test_import_pskc_encrypted_with_key(self, accounts_cli, tmp_path):
        """Test importing credentials from a key-encrypted PSKC file."""
        psk = "0123456789abcdef0123456789abcdef"
        pskc_file = tmp_path / "key_encrypted_import.pskc"

        # Export with encryption
        accounts_cli(
            "add",
            "key-import",
            "GEZDGNBVGY3TQOJQ",
            "-O",
            str(pskc_file),
            "--pskc-key",
            psk,
            "-f",
        )

        # Delete the credential
        accounts_cli("delete", "key-import", "-f")

        # Re-import with key
        accounts_cli("import", str(pskc_file), "-f", input=psk)
        creds = accounts_cli("list").output
        assert "key-import" in creds

    def test_export_hotp_credential(self, accounts_cli, tmp_path):
        """Test exporting an HOTP credential to PSKC."""
        pskc_file = tmp_path / "hotp_export.pskc"
        accounts_cli(
            "add",
            "hotp-export",
            "GEZDGNBVGY3TQOJQ",
            "-o",
            "HOTP",
            "-c",
            "10",
            "-O",
            str(pskc_file),
            "-f",
        )
        assert pskc_file.exists()

        content = pskc_file.read_text()
        assert "hotp" in content.lower()

    def test_export_totp_with_period(self, accounts_cli, tmp_path):
        """Test exporting a TOTP credential with custom period."""
        pskc_file = tmp_path / "totp_period.pskc"
        accounts_cli(
            "add",
            "totp-period",
            "GEZDGNBVGY3TQOJQ",
            "-P",
            "60",
            "-O",
            str(pskc_file),
            "-f",
        )
        assert pskc_file.exists()

        content = pskc_file.read_text()
        assert "60" in content

    def test_export_with_algorithm(self, accounts_cli, tmp_path):
        """Test exporting a credential with SHA256 algorithm."""
        pskc_file = tmp_path / "sha256_export.pskc"
        accounts_cli(
            "add",
            "sha256-test",
            "GEZDGNBVGY3TQOJQ",
            "-a",
            "SHA256",
            "-O",
            str(pskc_file),
            "-f",
        )
        assert pskc_file.exists()

        content = pskc_file.read_text()
        assert "SHA256" in content

    def test_export_with_digits(self, accounts_cli, tmp_path):
        """Test exporting a credential with 8 digits."""
        pskc_file = tmp_path / "8digits_export.pskc"
        accounts_cli(
            "add",
            "8digits-test",
            "GEZDGNBVGY3TQOJQ",
            "-d",
            "8",
            "-O",
            str(pskc_file),
            "-f",
        )
        assert pskc_file.exists()

        content = pskc_file.read_text()
        assert "8" in content



import pytest

from yubikit.core import TRANSPORT, YUBIKEY
from yubikit.management import CAPABILITY, FORM_FACTOR

from .. import condition

VALID_LOCK_CODE = "a" * 32
INVALID_LOCK_CODE_NON_HEX = "z" * 32


def _fido_only(capabilities):
    return capabilities & ~(CAPABILITY.U2F | CAPABILITY.FIDO2) == 0


def not_sky(device, info):
    if info.is_sky:
        return False
    if device.transport == TRANSPORT.NFC:
        return not (
            info.serial is None
            and _fido_only(info.supported_capabilities[TRANSPORT.USB])
        )
    else:
        return device.pid.yubikey_type != YUBIKEY.SKY


def not_mpe(info):
    if info.form_factor in (FORM_FACTOR.USB_A_BIO, FORM_FACTOR.USB_C_BIO):
        if info.supported_capabilities[TRANSPORT.USB] & CAPABILITY.PIV:
            return False
    return True


class TestConfigUSB:
    @pytest.fixture(autouse=True)
    @condition.check(not_sky, "SKY")
    @condition.check(not_mpe, "Bio MPE")
    @condition.min_version(5)
    def enable_all(self, ykman_cli, await_reboot):
        ykman_cli("config", "usb", "--enable-all", "-f")
        await_reboot()
        yield None
        ykman_cli("config", "usb", "--enable-all", "-f")
        await_reboot()

    @condition.capability(CAPABILITY.OTP, TRANSPORT.USB)
    def test_disable_otp(self, ykman_cli, await_reboot):
        ykman_cli("config", "usb", "--disable", "OTP", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "OTP" not in output

    @condition.capability(CAPABILITY.U2F, TRANSPORT.USB)
    def test_disable_u2f(self, ykman_cli, await_reboot):
        ykman_cli("config", "usb", "--disable", "U2F", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "FIDO U2F" not in output

    @condition.capability(CAPABILITY.OPENPGP, TRANSPORT.USB)
    def test_disable_openpgp(self, ykman_cli, await_reboot):
        ykman_cli("config", "usb", "--disable", "OPENPGP", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "OpenPGP" not in output

    @condition.capability(CAPABILITY.OPENPGP, TRANSPORT.USB)
    def test_disable_openpgp_alternative_syntax(self, ykman_cli, await_reboot):
        ykman_cli("config", "usb", "--disable", "openpgp", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "OpenPGP" not in output

    @condition.capability(CAPABILITY.PIV, TRANSPORT.USB)
    def test_disable_piv(self, ykman_cli, await_reboot):
        ykman_cli("config", "usb", "--disable", "PIV", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "PIV" not in output

    @condition.capability(CAPABILITY.OATH, TRANSPORT.USB)
    def test_disable_oath(self, ykman_cli, await_reboot):
        ykman_cli("config", "usb", "--disable", "OATH", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "OATH" not in output

    @condition.capability(CAPABILITY.FIDO2, TRANSPORT.USB)
    def test_disable_fido2(self, ykman_cli, await_reboot):
        ykman_cli("config", "usb", "--disable", "FIDO2", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "FIDO2" not in output

    @condition.capability(CAPABILITY.FIDO2, TRANSPORT.USB)
    def test_disable_and_enable(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("config", "usb", "--disable", "FIDO2", "--enable", "FIDO2", "-f")
        with pytest.raises(SystemExit):
            ykman_cli("config", "usb", "--enable-all", "--disable", "FIDO2", "-f")

    def test_disable_all(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli(
                "config",
                "usb",
                "-d",
                "FIDO2",
                "-d",
                "U2F",
                "-d",
                "OATH",
                "-d",
                "OPENPGP",
                "-d",
                "PIV",
                "-d",
                "OTP",
            )

    @condition.capability(CAPABILITY.U2F, TRANSPORT.USB)
    def test_mode_command(self, ykman_cli, await_reboot):
        ykman_cli("config", "mode", "ccid", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "FIDO U2F" not in output
        assert "FIDO2" not in output
        assert "OTP" not in output

        ykman_cli("config", "mode", "otp", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "FIDO U2F" not in output
        assert "FIDO2" not in output
        assert "OpenPGP" not in output
        assert "PIV" not in output
        assert "OATH" not in output

        ykman_cli("config", "mode", "fido", "-f")
        await_reboot()
        output = ykman_cli("config", "usb", "--list").output
        assert "OTP" not in output
        assert "OATH" not in output
        assert "PIV" not in output
        assert "OpenPGP" not in output


class TestConfigNFC:
    @pytest.fixture(autouse=True)
    @condition.check(not_sky)
    @condition.min_version(5)
    @condition.has_transport(TRANSPORT.NFC)
    def enable_all_nfc(self, ykman_cli, await_reboot):
        ykman_cli("config", "nfc", "--enable-all", "-f")
        await_reboot()
        yield None
        ykman_cli("config", "nfc", "--enable-all", "-f")
        await_reboot()

    @condition.capability(CAPABILITY.OTP, TRANSPORT.NFC)
    def test_disable_otp(self, ykman_cli):
        ykman_cli("config", "nfc", "--disable", "OTP", "-f")
        output = ykman_cli("config", "nfc", "--list").output
        assert "OTP" not in output

    @condition.capability(CAPABILITY.U2F, TRANSPORT.NFC)
    def test_disable_u2f(self, ykman_cli):
        ykman_cli("config", "nfc", "--disable", "U2F", "-f")
        output = ykman_cli("config", "nfc", "--list").output
        assert "FIDO U2F" not in output

    @condition.capability(CAPABILITY.OPENPGP, TRANSPORT.NFC)
    def test_disable_openpgp(self, ykman_cli):
        ykman_cli("config", "nfc", "--disable", "OPENPGP", "-f")
        output = ykman_cli("config", "nfc", "--list").output
        assert "OpenPGP" not in output

    @condition.capability(CAPABILITY.PIV, TRANSPORT.NFC)
    def test_disable_piv(self, ykman_cli):
        ykman_cli("config", "nfc", "--disable", "PIV", "-f")
        output = ykman_cli("config", "nfc", "--list").output
        assert "PIV" not in output

    @condition.capability(CAPABILITY.OATH, TRANSPORT.NFC)
    def test_disable_oath(self, ykman_cli):
        ykman_cli("config", "nfc", "--disable", "OATH", "-f")
        output = ykman_cli("config", "nfc", "--list").output
        assert "OATH" not in output

    @condition.capability(CAPABILITY.FIDO2, TRANSPORT.NFC)
    def test_disable_fido2(self, ykman_cli):
        ykman_cli("config", "nfc", "--disable", "FIDO2", "-f")
        output = ykman_cli("config", "nfc", "--list").output
        assert "FIDO2" not in output

    @condition.transport(TRANSPORT.USB)
    def test_disable_all(self, ykman_cli):
        ykman_cli("config", "nfc", "--disable-all", "-f")
        output = ykman_cli("config", "nfc", "--list").output
        assert not output

    @condition.capability(CAPABILITY.FIDO2, TRANSPORT.NFC)
    def test_disable_and_enable(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("config", "nfc", "--disable", "FIDO2", "--enable", "FIDO2", "-f")
        with pytest.raises(SystemExit):
            ykman_cli("config", "nfc", "--disable-all", "--enable", "FIDO2", "-f")
        with pytest.raises(SystemExit):
            ykman_cli("config", "nfc", "--enable-all", "--disable", "FIDO2", "-f")
        with pytest.raises(SystemExit):
            ykman_cli("config", "nfc", "--enable-all", "--disable-all", "FIDO2", "-f")


class TestConfigLockCode:
    @pytest.fixture(autouse=True)
    @condition.min_version(5)
    def preconditions(self):
        pass

    def test_set_lock_code(self, ykman_cli):
        ykman_cli("config", "set-lock-code", "--new-lock-code", VALID_LOCK_CODE)
        output = ykman_cli("info").output
        assert "Configured capabilities are protected by a lock code" in output
        ykman_cli("config", "set-lock-code", "-l", VALID_LOCK_CODE, "--clear")
        output = ykman_cli("info").output
        assert "Configured capabilities are protected by a lock code" not in output

    def test_set_invalid_lock_code(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("config", "set-lock-code", "--new-lock-code", "aaaa")

        with pytest.raises(SystemExit):
            ykman_cli(
                "config", "set-lock-code", "--new-lock-code", INVALID_LOCK_CODE_NON_HEX
            )



import os
from enum import Enum

import pytest

from yubikit.core.smartcard import AID, SmartCardProtocol
from yubikit.core.smartcard.scp import KeyRef, Scp11KeyParams
from yubikit.securitydomain import SecurityDomainSession

from . import condition


@pytest.fixture
@condition.min_version(5, 7, 2)
def scp_params(ccid_connection):
    ref = KeyRef(0x13, 0x1)
    sd = SecurityDomainSession(ccid_connection)
    if ref not in sd.get_key_information():
        pytest.skip("Default SCP11b key not present, skipping SCP test")
    chain = sd.get_certificate_bundle(ref)
    return Scp11KeyParams(ref, chain[-1].public_key())


class Configuration(Enum):
    default = 0
    force_short = 1
    no_config = 2

    def __str__(self):
        return self.name


@pytest.mark.parametrize("size", (10, 255, 256, 512, 2048))
@pytest.mark.parametrize("config", Configuration)
def test_scp_apdus(config, size, ccid_connection, scp_params, version):
    protocol = SmartCardProtocol(ccid_connection)
    if config == Configuration.default:
        protocol.configure(version)
    elif config == Configuration.force_short:
        protocol.configure(version, True)
    else:
        pass  # no configuration

    protocol.select(AID.MANAGEMENT)
    protocol.init_scp(scp_params)

    payload = os.urandom(size)
    # ECHO command
    resp = protocol.send_apdu(0, 1, 0, 0, payload)
    assert resp == payload



# Copyright (c) 2015-2020 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import abc
from enum import Enum
from threading import Event
from typing import Callable, Hashable

from yubikit.core import PID, TRANSPORT, YubiKeyDevice


class REINSERT_STATUS(Enum):
    REMOVE = 1
    REINSERT = 2


class CancelledException(Exception):
    """Raised when the caller cancels an operation."""


class YkmanDevice(YubiKeyDevice):
    """YubiKey device reference, with optional PID"""

    def __init__(self, transport: TRANSPORT, fingerprint: Hashable, pid: PID | None):
        super(YkmanDevice, self).__init__(transport, fingerprint)
        self._pid = pid

    @property
    def pid(self) -> PID | None:
        """Return the PID of the YubiKey, if available."""
        return self._pid

    def reinsert(
        self,
        reinsert_cb: Callable[[REINSERT_STATUS], None] | None = None,
        event: Event | None = None,
    ) -> None:
        """Wait for the user to remove and reinsert the YubiKey.

        This may be required to perform certain operations, such as FIDO reset.

        This method will attempt to verify that the same YubiKey is reinserted,
        but it will only fail when this is definitely not the case (eg. if the serial
        number does not match).

        :param reinsert_cb: Callback to indicate the the YubiKey has been removed,
        and should be reinserted.
        :param event: Optional event to cancel (throws CancelledException).
        """
        self._do_reinsert(reinsert_cb or (lambda _: None), event or Event())

    @abc.abstractmethod
    def _do_reinsert(
        self, reinsert_cb: Callable[[REINSERT_STATUS], None], event: Event
    ) -> None:
        pass

    def __repr__(self):
        return "%s(pid=%04x, fingerprint=%r)" % (
            type(self).__name__,
            self.pid or 0,
            self.fingerprint,
        )



import pytest

from yubikit.core import TRANSPORT
from yubikit.core.smartcard import AID, SW, ApduError
from yubikit.management import CAPABILITY
from yubikit.oath import (
    HASH_ALGORITHM,
    OATH_TYPE,
    CredentialData,
    OathSession,
)

from . import condition

KEY = bytes.fromhex("01020304050607080102030405060708")


@pytest.fixture
@condition.capability(CAPABILITY.OATH)
def session(ccid_connection, info, scp_params):
    fips = CAPABILITY.OATH in info.fips_capable
    if ccid_connection.transport == TRANSPORT.NFC and fips:
        oath = OathSession(ccid_connection, scp_params)
    else:
        oath = OathSession(ccid_connection)
    oath.reset()

    if fips:
        oath.set_key(KEY)

    yield oath


CRED_DATA = CredentialData("name", OATH_TYPE.TOTP, HASH_ALGORITHM.SHA1, b"secret")


class TestFunctions:
    @condition.min_version(5, 3)
    def test_rename(self, session):
        cred = session.put_credential(CRED_DATA)
        new_id = session.rename_credential(cred.id, "newname", "newissuer")
        with pytest.raises(ApduError):
            session.calculate(cred.id, b"challenge")
        session.calculate(new_id, b"challenge")

    @condition.min_version(5, 3)
    def test_rename_to_existing(self, session):
        cred = session.put_credential(CRED_DATA)
        new_id = session.rename_credential(cred.id, "newname", "newissuer")
        with pytest.raises(ApduError):
            session.rename_credential(new_id, "newname", "newissuer")


class TestLockPreventsAccess:
    @pytest.fixture(autouse=True)
    def set_lock(self, session):
        assert not session.locked
        session.put_credential(CRED_DATA)
        session.set_key(KEY)

        # Force re-select to lock
        session.protocol.connection.connection.disconnect()
        session.protocol.connection.connection.connect()
        session.protocol.select(AID.OATH)

    def test_list(self, session):
        with pytest.raises(ApduError) as ctx:
            session.list_credentials()
        assert ctx.value.sw == SW.SECURITY_CONDITION_NOT_SATISFIED

    def test_calculate(self, session):
        with pytest.raises(ApduError) as ctx:
            session.calculate(CRED_DATA.get_id(), b"challenge")
        assert ctx.value.sw == SW.SECURITY_CONDITION_NOT_SATISFIED

    def test_calculate_all(self, session):
        with pytest.raises(ApduError) as ctx:
            session.calculate_all()
        assert ctx.value.sw == SW.SECURITY_CONDITION_NOT_SATISFIED

    def test_delete(self, session):
        with pytest.raises(ApduError) as ctx:
            session.delete_credential(CRED_DATA.get_id())
        assert ctx.value.sw == SW.SECURITY_CONDITION_NOT_SATISFIED

    @condition.min_version(5, 3)
    def test_rename(self, session):
        with pytest.raises(ApduError) as ctx:
            session.rename_credential(CRED_DATA.get_id(), "renamed")
        assert ctx.value.sw == SW.SECURITY_CONDITION_NOT_SATISFIED


HMAC_VECTORS = {
    b"\x0b" * 20: {
        b"Hi There": {
            HASH_ALGORITHM.SHA256: bytes.fromhex(
                "b0344c61d8db38535ca8afceaf0bf12b881dc200c9833da726e9376c2e32cff7"
            ),
            HASH_ALGORITHM.SHA512: bytes.fromhex(
                "87aa7cdea5ef619d4ff0b4241a1d6cb02379f4e2ce4ec2787ad0b30545e17cde"
                "daa833b7d6b8a702038b274eaea3f4e4be9d914eeb61f1702e696c203a126854"
            ),
        }
    },
    b"Jefe": {
        b"what do ya want for nothing?": {
            HASH_ALGORITHM.SHA256: bytes.fromhex(
                "5bdcc146bf60754e6a042426089575c75a003f089d2739839dec58b964ec3843"
            ),
            HASH_ALGORITHM.SHA512: bytes.fromhex(
                "164b7a7bfcf819e2e395fbe73b56e0a387bd64222e831fd610270cd7ea250554"
                "9758bf75c05a994a6d034f65f8f0e6fdcaeab1a34d4a6b4b636e070a38bce737"
            ),
        }
    },
    b"\xaa" * 20: {
        b"\xdd" * 50: {
            HASH_ALGORITHM.SHA256: bytes.fromhex(
                "773ea91e36800e46854db8ebd09181a72959098b3ef8c122d9635514ced565fe"
            ),
            HASH_ALGORITHM.SHA512: bytes.fromhex(
                "fa73b0089d56a284efb0f0756c890be9b1b5dbdd8ee81a3655f83e33b2279d39"
                "bf3e848279a722c806b485a47e67c807b946a337bee8942674278859e13292fb"
            ),
        }
    },
    bytes.fromhex("0102030405060708090a0b0c0d0e0f10111213141516171819"): {
        b"\xcd" * 50: {
            HASH_ALGORITHM.SHA256: bytes.fromhex(
                "82558a389a443c0ea4cc819899f2083a85f0faa3e578f8077a2e3ff46729665b"
            ),
            HASH_ALGORITHM.SHA512: bytes.fromhex(
                "b0ba465637458c6990e5a8c5f61d4af7e576d97ff94b872de76f8050361ee3db"
                "a91ca5c11aa25eb4d679275cc5788063a5f19741120c4f2de2adebeb10a298dd"
            ),
        }
    },
}


HMAC_PARAMS = [
    (key, timestamp, algo, HMAC_VECTORS[key][timestamp][algo])
    for key in HMAC_VECTORS
    for timestamp in HMAC_VECTORS[key]
    for algo in HMAC_VECTORS[key][timestamp]
]


def _ids_hmac(params):
    key, challenge, hash_algorithm, expected = params
    key_s = key.hex() if len(key) < 6 else key[:6].hex() + "..."
    challenge_s = challenge.hex() if len(challenge) < 6 else challenge[:6].hex() + "..."
    return f"{hash_algorithm.name}-{key_s}-{challenge_s}"


class TestHmacVectors:
    @pytest.mark.parametrize("params", HMAC_PARAMS, ids=_ids_hmac)
    def test_vector(self, info, session, params):
        key, challenge, hash_algorithm, expected = params
        if hash_algorithm == HASH_ALGORITHM.SHA512:
            if info.version[0] <= 4:
                if info.is_fips or info.version < (4, 3, 1):
                    pytest.skip("SHA512 requires (non-FIPS) YubiKey 4.3.1 or later")
        cred = session.put_credential(
            CredentialData("test", OATH_TYPE.TOTP, hash_algorithm, key)
        )
        value = session.calculate(cred.id, challenge)
        assert value == expected


TOTP_VECTOR_KEYS = {
    HASH_ALGORITHM.SHA1: b"12345678901234567890",
    HASH_ALGORITHM.SHA256: b"12345678901234567890123456789012",
    HASH_ALGORITHM.SHA512: b"12345678901234567890123456789012"
    b"34567890123456789012345678901234",
}
TOTP_VECTORS = {
    59: {
        HASH_ALGORITHM.SHA1: "94287082",
        HASH_ALGORITHM.SHA256: "46119246",
        HASH_ALGORITHM.SHA512: "90693936",
    },
    1111111109: {
        HASH_ALGORITHM.SHA1: "07081804",
        HASH_ALGORITHM.SHA256: "68084774",
        HASH_ALGORITHM.SHA512: "25091201",
    },
}


TOTP_PARAMS = [
    (timestamp, algo, TOTP_VECTORS[timestamp][algo], TOTP_VECTOR_KEYS[algo])
    for timestamp in TOTP_VECTORS
    for algo in TOTP_VECTORS[timestamp]
]


class TestTotpVectors:
    @pytest.mark.parametrize("digits", [6, 8])
    @pytest.mark.parametrize(
        "params", TOTP_PARAMS, ids=lambda x: "{1.name}-{0}".format(*x)
    )
    def test_vector(self, info, session, params, digits):
        timestamp, hash_algorithm, value, key = params
        if hash_algorithm == HASH_ALGORITHM.SHA512:
            if info.version[0] <= 4:
                if info.is_fips or info.version < (4, 3, 1):
                    pytest.skip("SHA512 requires (non-FIPS) YubiKey 4.3.1 or later")

        cred = session.put_credential(
            CredentialData("test", OATH_TYPE.TOTP, hash_algorithm, key, digits)
        )
        code = session.calculate_code(cred, timestamp)
        assert len(code.value) == digits
        assert value.endswith(code.value)


HOTP_VECTORS = {
    b"12345678901234567890": [
        "84755224",
        "94287082",
        "37359152",
        "26969429",
        "40338314",
        "68254676",
        "18287922",
        "82162583",
        "73399871",
        "45520489",
    ]
}


class TestHotpVectors:
    @pytest.mark.parametrize("digits", [6, 8])
    @pytest.mark.parametrize(
        "params", HOTP_VECTORS.items(), ids=lambda x: "{0}".format(*x)
    )
    def test_vector(self, session, params, digits):
        key, values = params

        cred = session.put_credential(
            CredentialData("test", OATH_TYPE.HOTP, HASH_ALGORITHM.SHA1, key, digits)
        )
        for expected in values:
            code = session.calculate_code(cred)
            assert len(code.value) == digits
            assert expected.endswith(code.value)



#  vim: set fileencoding=utf-8 :

# Copyright (c) 2018 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re
from time import sleep

import pytest

from yubikit.management import CAPABILITY

from .. import condition


def no_pin_complexity(info):
    """PIN complexity enabled"""
    return not info.pin_complexity


@pytest.fixture(autouse=True)
@condition.capability(CAPABILITY.OTP)
def ensure_otp():
    pass


class TestSlotStatus:
    def test_ykman_otp_info(self, ykman_cli):
        info = ykman_cli("otp", "info").output
        assert "Slot 1:" in info
        assert "Slot 2:" in info

    @condition.check(no_pin_complexity)
    def test_ykman_swap_slots(self, ykman_cli):
        info = ykman_cli("otp", "info").output
        if "programmed" not in info:
            ykman_cli("otp", "static", "2", "incredible")
        output = ykman_cli("otp", "swap", "-f").output
        assert "Slots swapped" in output
        output = ykman_cli("otp", "swap", "-f").output
        assert "Slots swapped" in output

    @condition.yk4_fips(False)
    def test_ykman_otp_info_does_not_indicate_fips_mode_for_non_fips_key(
        self, ykman_cli
    ):  # noqa: E501
        info = ykman_cli("otp", "info").output
        assert "FIPS Approved Mode:" not in info


class TestReclaimTimeout:
    @condition.check(no_pin_complexity)
    def test_update_after_reclaim(self, ykman_cli):
        info = ykman_cli("otp", "info").output
        if "programmed" not in info:
            ykman_cli("otp", "static", "2", "incredible")
        ykman_cli("otp", "swap", "-f")
        ykman_cli("otp", "swap", "-f")
        sleep(4)  # Ensure reclaim
        ykman_cli("otp", "swap", "-f")
        ykman_cli("otp", "swap", "-f")


class TestSlotStaticPassword:
    @pytest.fixture(autouse=True)
    @condition.check(no_pin_complexity)
    def delete_slot(self, ykman_cli):
        try:
            ykman_cli("otp", "delete", "2", "-f")
        except SystemExit:
            pass
        yield None
        try:
            ykman_cli("otp", "delete", "2", "-f")
        except SystemExit:
            pass

    def test_too_long(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("otp", "static", "2", "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")

    def test_unsupported_chars(self, ykman_cli):
        with pytest.raises(ValueError):
            ykman_cli("otp", "static", "2", "ö")
        with pytest.raises(ValueError):
            ykman_cli("otp", "static", "2", "@")

    def test_provide_valid_pw(self, ykman_cli):
        ykman_cli("otp", "static", "2", "higngdukgerjktbbikrhirngtlkkttkb")
        assert "Slot 2: programmed" in ykman_cli("otp", "info").output

    def test_provide_valid_pw_prompt(self, ykman_cli):
        ykman_cli("otp", "static", "2", input="higngdukgerjktbbikrhirngtlkkttkb\ny\n")
        assert "Slot 2: programmed" in ykman_cli("otp", "info").output

    def test_generate_pw_too_long(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("otp", "static", "2", "--generate", "--length", "39")

    def test_generate_pw_blank_length(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("otp", "static", "2", "--generate", "--length")

    def test_generate_zero_length(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("otp", "static", "2", "--generate", "--length", "0")

    def test_generate_pw(self, ykman_cli):
        ykman_cli("otp", "static", "2", "--generate", "--length", "38")
        assert "Slot 2: programmed" in ykman_cli("otp", "info").output

    def test_generate_pw_default_length(self, ykman_cli):
        ykman_cli("otp", "static", "2", "--generate")
        assert "Slot 2: programmed" in ykman_cli("otp", "info").output

    def test_us_scancodes(self, ykman_cli):
        ykman_cli("otp", "static", "2", "abcABC123", "--keyboard-layout", "US")
        ykman_cli("otp", "static", "2", "@!)", "-f", "--keyboard-layout", "US")

    def test_de_scancodes(self, ykman_cli):
        ykman_cli("otp", "static", "2", "abcABC123", "--keyboard-layout", "DE")
        ykman_cli("otp", "static", "2", "Üßö", "-f", "--keyboard-layout", "DE")

    def test_overwrite_prompt(self, ykman_cli):
        ykman_cli("otp", "static", "2", "bbb")
        with pytest.raises(SystemExit):
            ykman_cli("otp", "static", "2", "ccc")
        ykman_cli("otp", "static", "2", "ddd", "-f")
        assert "Slot 2: programmed" in ykman_cli("otp", "info").output


class TestSlotProgramming:
    @pytest.fixture(autouse=True)
    @condition.check(no_pin_complexity)
    def delete_slot(self, ykman_cli):
        try:
            ykman_cli("otp", "delete", "2", "-f")
        except SystemExit:
            pass
        yield None
        try:
            ykman_cli("otp", "delete", "2", "-f")
        except SystemExit:
            pass

    def test_ykman_program_otp_slot_2(self, ykman_cli):
        ykman_cli(
            "otp",
            "yubiotp",
            "2",
            "--public-id",
            "vvccccfiluij",
            "--private-id",
            "267e0a88949b",
            "--key",
            "b8e31ab90bb8830e3c1fe1b483a8e0d4",
            "-f",
        )
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_otp_slot_2_prompt(self, ykman_cli):
        ykman_cli(
            "otp",
            "yubiotp",
            "2",
            input="vvccccfiluij\n267e0a88949b\nb8e31ab90bb8830e3c1fe1b483a8e0d4\ny\n",
        )
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_otp_slot_2_options(self, ykman_cli):
        output = ykman_cli(
            "otp",
            "yubiotp",
            "2",
            "--public-id",
            "vvccccfiluij",
            "--private-id",
            "267e0a88949b",
            "--key",
            "b8e31ab90bb8830e3c1fe1b483a8e0d4",
            "-f",
        ).output
        assert "" == output
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_otp_slot_2_generated_all(self, ykman_cli):
        output = ykman_cli(
            "otp",
            "yubiotp",
            "2",
            "-f",
            "--serial-public-id",
            "--generate-private-id",
            "--generate-key",
        ).output
        assert "Using YubiKey serial as public ID" in output
        assert "Using a randomly generated private ID" in output
        assert "Using a randomly generated secret key" in output
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_otp_slot_2_serial_public_id(self, ykman_cli):
        output = ykman_cli(
            "otp",
            "yubiotp",
            "2",
            "--serial-public-id",
            "--private-id",
            "267e0a88949b",
            "--key",
            "b8e31ab90bb8830e3c1fe1b483a8e0d4",
            "-f",
        ).output
        assert "Using YubiKey serial as public ID" in output
        assert "generated private ID" not in output
        assert "generated secret key" not in output
        self._check_slot_2_programmed(ykman_cli)

    def test_invalid_public_id(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("otp", "yubiotp", "-P", "imnotmodhex!")

    def test_ykman_program_otp_slot_2_generated_private_id(self, ykman_cli):
        output = ykman_cli(
            "otp",
            "yubiotp",
            "2",
            "--public-id",
            "vvccccfiluij",
            "--generate-private-id",
            "--key",
            "b8e31ab90bb8830e3c1fe1b483a8e0d4",
            "-f",
        ).output
        assert "serial as public ID" not in output
        assert "Using a randomly generated private ID" in output
        assert "generated secret key" not in output
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_otp_slot_2_generated_secret_key(self, ykman_cli):
        output = ykman_cli(
            "otp",
            "yubiotp",
            "2",
            "--public-id",
            "vvccccfiluij",
            "--private-id",
            "267e0a88949b",
            "--generate-key",
            "-f",
        ).output
        assert "serial as public ID" not in output
        assert "generated private ID" not in output
        assert "Using a randomly generated secret key" in output
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_otp_slot_2_serial_id_conflicts_public_id(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli(
                "otp",
                "yubiotp",
                "2",
                "-f",
                "--serial-public-id",
                "--public-id",
                "vvccccfiluij",
                "--generate-private-id",
                "--generate-key",
            )
        self._check_slot_2_not_programmed(ykman_cli)

    def test_ykman_program_otp_slot_2_generate_id_conflicts_private_id(self, ykman_cli):  # noqa: E501
        with pytest.raises(SystemExit):
            ykman_cli(
                "otp",
                "yubiotp",
                "2",
                "-f",
                "--serial-public-id",
                "--generate-private-id",
                "--private-id",
                "267e0a88949b",
                "--generate-key",
            )
        self._check_slot_2_not_programmed(ykman_cli)

    def test_ykman_program_otp_slot_2_generate_key_conflicts_key(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli(
                "otp",
                "yubiotp",
                "2",
                "-f",
                "--serial-public-id",
                "--generate-private-id",
                "--generate-key",
                "--key",
                "b8e31ab90bb8830e3c1fe1b483a8e0d4",
            )
        self._check_slot_2_not_programmed(ykman_cli)

    def test_ykman_program_chalresp_slot_2(self, ykman_cli):
        ykman_cli("otp", "chalresp", "2", "abba", "-f")
        self._check_slot_2_programmed(ykman_cli)
        ykman_cli("otp", "chalresp", "2", "--totp", "abba", "-f")
        self._check_slot_2_programmed(ykman_cli)
        ykman_cli("otp", "chalresp", "2", "--touch", "abba", "-f")
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_chalresp_slot_2_force_fails_without_key(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("otp", "chalresp", "2", "-f")
        self._check_slot_2_not_programmed(ykman_cli)

    def test_ykman_program_chalresp_slot_2_generated(self, ykman_cli):
        output = ykman_cli("otp", "chalresp", "2", "-f", "-g").output
        assert re.search(
            r"Using a randomly generated key \(hex\): [0-9a-f]{40}", output
        )
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_chalresp_slot_2_generated_fails_if_also_given(
        self, ykman_cli
    ):  # noqa: E501
        with pytest.raises(SystemExit):
            ykman_cli("otp", "chalresp", "2", "-f", "-g", "abababab")

    def test_ykman_program_chalresp_slot_2_prompt(self, ykman_cli):
        ykman_cli("otp", "chalresp", "2", input="abba\ny\n")
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_hotp_slot_2(self, ykman_cli):
        ykman_cli("otp", "hotp", "2", "27KIZZE3SD7GE2FVJJBAXEI3I6RRTPGM", "-f")
        self._check_slot_2_programmed(ykman_cli)

    def test_ykman_program_hotp_slot_2_prompt(self, ykman_cli):
        ykman_cli("otp", "hotp", "2", input="abba\ny\n")
        self._check_slot_2_programmed(ykman_cli)

    def test_update_settings_enter_slot_2(self, ykman_cli):
        ykman_cli("otp", "static", "2", "-f", "-g", "-l", "20")
        output = ykman_cli("otp", "settings", "2", "-f", "--no-enter").output
        assert "updated" in output

    def test_delete_slot_2(self, ykman_cli):
        ykman_cli("otp", "static", "2", "-f", "-g", "-l", "20")
        output = ykman_cli("otp", "delete", "2", "-f").output
        assert "deleted" in output
        status = ykman_cli("otp", "info").output
        assert "Slot 2: empty" in status

    def test_access_code_slot_2(self, ykman_cli):
        ykman_cli(
            "otp",
            "--access-code",
            "111111111111",
            "static",
            "2",
            "--generate",
            "--length",
            "10",
        )
        self._check_slot_2_programmed(ykman_cli)
        self._check_slot_2_has_access_code(ykman_cli)
        ykman_cli("otp", "--access-code", "111111111111", "delete", "2", "-f")
        status = ykman_cli("otp", "info").output
        assert "Slot 2: empty" in status

    @condition.min_version(4, 3, 2)
    @condition.max_version(4, 3, 5)
    def test_update_access_code_fails_on_yk_432_to_435(self, ykman_cli):
        ykman_cli("otp", "static", "2", "--generate", "--length", "10")

        self._check_slot_2_programmed(ykman_cli)

        with pytest.raises(SystemExit):
            ykman_cli("otp", "settings", "--new-access-code", "111111111111", "2", "-f")

        ykman_cli(
            "otp",
            "--access-code",
            "111111111111",
            "static",
            "2",
            "-f",
            "--generate",
            "--length",
            "10",
        )

        with pytest.raises(SystemExit):
            ykman_cli("otp", "delete", "2", "-f")

        with pytest.raises(SystemExit):
            ykman_cli(
                "otp",
                "--access-code",
                "111111111111",
                "settings",
                "--new-access-code",
                "222222222222",
                "2",
                "-f",
            )

        ykman_cli("otp", "--access-code", "111111111111", "delete", "2", "-f")

    @condition.min_version(4, 3, 2)
    @condition.max_version(4, 3, 5)
    def test_delete_access_code_fails_on_yk_432_to_435(self, ykman_cli):
        ykman_cli(
            "otp",
            "--access-code",
            "111111111111",
            "static",
            "2",
            "--generate",
            "--length",
            "10",
        )

        self._check_slot_2_programmed(ykman_cli)

        with pytest.raises(SystemExit):
            ykman_cli(
                "otp",
                "--access-code",
                "111111111111",
                "settings",
                "--delete-access-code",
                "2",
                "-f",
            )

        with pytest.raises(SystemExit):
            ykman_cli("otp", "delete", "2", "-f")

        ykman_cli("otp", "--access-code", "111111111111", "delete", "2", "-f")

    @condition.check(lambda version: not (4, 3, 2) <= version <= (4, 3, 5))
    def test_update_access_code_slot_2(self, ykman_cli):
        ykman_cli("otp", "static", "2", "--generate", "--length", "10")

        self._check_slot_2_programmed(ykman_cli)
        self._check_slot_2_does_not_have_access_code(ykman_cli)

        ykman_cli("otp", "settings", "--new-access-code", "111111111111", "2", "-f")
        self._check_slot_2_has_access_code(ykman_cli)

        ykman_cli(
            "otp",
            "--access-code",
            "111111111111",
            "settings",
            "--delete-access-code",
            "2",
            "-f",
        )
        self._check_slot_2_does_not_have_access_code(ykman_cli)

        ykman_cli("otp", "delete", "2", "-f")

    @condition.check(lambda version: not (4, 3, 2) <= version <= (4, 3, 5))
    def test_update_access_code_prompt_slot_2(self, ykman_cli):
        ykman_cli("otp", "static", "2", "--generate", "--length", "10")

        self._check_slot_2_programmed(ykman_cli)
        self._check_slot_2_does_not_have_access_code(ykman_cli)

        ykman_cli(
            "otp", "settings", "--new-access-code", "-", "2", "-f", input="111111111111"
        )
        self._check_slot_2_has_access_code(ykman_cli)

        ykman_cli(
            "otp",
            "--access-code",
            "-",
            "settings",
            "--delete-access-code",
            "2",
            "-f",
            input="111111111111",
        )
        self._check_slot_2_does_not_have_access_code(ykman_cli)

        ykman_cli("otp", "delete", "2", "-f")

    @condition.check(lambda version: not (4, 3, 2) <= version <= (4, 3, 5))
    def test_new_access_code_conflicts_with_delete_access_code(self, ykman_cli):
        ykman_cli("otp", "static", "2", "--generate", "--length", "10")

        self._check_slot_2_programmed(ykman_cli)
        self._check_slot_2_does_not_have_access_code(ykman_cli)

        with pytest.raises(SystemExit):
            ykman_cli(
                "otp",
                "settings",
                "--delete-access-code",
                "--new-access-code",
                "111111111111",
                "2",
                "-f",
            )
        self._check_slot_2_does_not_have_access_code(ykman_cli)

        ykman_cli("otp", "settings", "--new-access-code", "111111111111", "2", "-f")

        with pytest.raises(SystemExit):
            ykman_cli(
                "otp",
                "settings",
                "--delete-access-code",
                "--new-access-code",
                "111111111111",
                "2",
                "-f",
            )
        self._check_slot_2_has_access_code(ykman_cli)

        ykman_cli("otp", "--access-code", "111111111111", "delete", "2", "-f")

    def _check_slot_2_programmed(self, ykman_cli):
        status = ykman_cli("otp", "info").output
        assert "Slot 2: programmed" in status

    def _check_slot_2_not_programmed(self, ykman_cli):
        status = ykman_cli("otp", "info").output
        assert "Slot 2: empty" in status

    def _check_slot_2_has_access_code(self, ykman_cli):
        with pytest.raises(SystemExit):
            ykman_cli("otp", "settings", "--pacing", "0", "2", "-f")

        ykman_cli(
            "otp",
            "--access-code",
            "111111111111",
            "settings",
            "--pacing",
            "0",
            "2",
            "-f",
        )

    def _check_slot_2_does_not_have_access_code(self, ykman_cli):
        ykman_cli("otp", "settings", "--pacing", "0", "2", "-f")


class TestSlotCalculate:
    @pytest.fixture(autouse=True)
    @condition.check(no_pin_complexity)
    def delete_slot(self, ykman_cli):
        try:
            ykman_cli("otp", "delete", "2", "-f")
        except SystemExit:
            pass
        yield None
        try:
            ykman_cli("otp", "delete", "2", "-f")
        except SystemExit:
            pass

    def test_calculate_hex(self, ykman_cli):
        ykman_cli("otp", "chalresp", "2", "abba", "-f")
        output = ykman_cli("otp", "calculate", "2", "abba").output
        assert "f8de2586056d89d8b961a072d1245a495d2155e1" in output

    def test_calculate_totp(self, ykman_cli):
        ykman_cli("otp", "chalresp", "2", "abba", "-f")
        output = ykman_cli("otp", "calculate", "2", "999", "-T").output
        assert "533486" == output.strip()
        output = ykman_cli("otp", "calculate", "2", "999", "-T", "-d", "8").output
        assert "04533486" == output.strip()
        output = ykman_cli("otp", "calculate", "2", "-T").output
        assert 6 == len(output.strip())
        output = ykman_cli("otp", "calculate", "2", "-T", "-d", "8").output
        assert 8 == len(output.strip())


class TestFipsMode:
    @pytest.fixture(autouse=True)
    @condition.yk4_fips(True)
    def delete_slots(self, ykman_cli):
        try:
            ykman_cli("otp", "delete", "1", "-f")
        except SystemExit:
            pass
        try:
            ykman_cli("otp", "delete", "2", "-f")
        except SystemExit:
            pass
        yield None

    def test_not_fips_mode_if_no_slot_programmed(self, ykman_cli):
        info = ykman_cli("otp", "info").output
        assert "FIPS Approved Mode: No" in info

    def test_not_fips_mode_if_slot_1_not_programmed(self, ykman_cli):
        ykman_cli("otp", "static", "2", "--generate", "--length", "10")

        info = ykman_cli("otp", "info").output
        assert "FIPS Approved Mode: No" in info

    def test_not_fips_mode_if_slot_2_not_programmed(self, ykman_cli):
        ykman_cli("otp", "static", "1", "--generate", "--length", "10")

        info = ykman_cli("otp", "info").output
        assert "FIPS Approved Mode: No" in info

    def test_not_fips_mode_if_no_slot_has_access_code(self, ykman_cli):
        ykman_cli("otp", "static", "1", "--generate", "--length", "10")
        ykman_cli("otp", "static", "2", "--generate", "--length", "10")

        info = ykman_cli("otp", "info").output
        assert "FIPS Approved Mode: No" in info

    def test_not_fips_mode_if_only_slot_1_has_access_code(self, ykman_cli):
        ykman_cli("otp", "static", "1", "--generate", "--length", "10")
        ykman_cli("otp", "static", "2", "--generate", "--length", "10")

        ykman_cli("otp", "settings", "--new-access-code", "111111111111", "1", "-f")

        info = ykman_cli("otp", "info").output
        assert "FIPS Approved Mode: No" in info

        ykman_cli("otp", "--access-code", "111111111111", "delete", "1", "-f")

    def test_not_fips_mode_if_only_slot_2_has_access_code(self, ykman_cli):
        ykman_cli("otp", "static", "1", "--generate", "--length", "10")
        ykman_cli("otp", "static", "2", "--generate", "--length", "10")

        ykman_cli("otp", "settings", "--new-access-code", "111111111111", "2", "-f")

        info = ykman_cli("otp", "info").output
        assert "FIPS Approved Mode: No" in info

        ykman_cli("otp", "--access-code", "111111111111", "delete", "2", "-f")

    def test_fips_mode_if_both_slots_have_access_code(self, ykman_cli):
        ykman_cli("otp", "static", "--generate", "--length", "10", "1", "-f")
        ykman_cli("otp", "static", "--generate", "--length", "10", "2", "-f")

        ykman_cli("otp", "settings", "--new-access-code", "111111111111", "1", "-f")
        ykman_cli("otp", "settings", "--new-access-code", "111111111111", "2", "-f")

        info = ykman_cli("otp", "info").output
        assert "FIPS Approved Mode: Yes" in info

        ykman_cli("otp", "--access-code", "111111111111", "delete", "1", "-f")
        ykman_cli("otp", "--access-code", "111111111111", "delete", "2", "-f")
